searchData={"items":[{"type":"module","title":"EventStreamex","doc":"The event streaming system supervisor.\n\nThis supervisor is the one started by `EventStreamex.Orchestrator` when\nthe current node is the master node.\n\nThis supervisor is responsible for starting the whole system.","ref":"EventStreamex.html"},{"type":"function","title":"EventStreamex.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.html#child_spec/1"},{"type":"function","title":"EventStreamex.restart_scheduler/0","doc":"Restarts the `EventStreamex.Operators.Scheduler`.\n\nShould only be used when the Scheduler crashes because of an operator\nthat fails after several restarts.\n\nBecause this failure is critical, the sheduler will fail and must be restarted\nmanually using this function when the error is solved with the operator.\n\nWhen the scheduler is restarted, the failed operator will be restarted\nand then, the rest of the events queue will be executed.","ref":"EventStreamex.html#restart_scheduler/0"},{"type":"module","title":"EventStreamex.EventListener","doc":"Listens for new database WAL events in a `Phoenix.LiveView`.\n\nAs soon as something happens in the database (insert, delete, update),\na WAL event is sent and dispatched to operators that listen to it, and to\nlive views that uses the `EventStreamex.EventListener` module.\n\nLet's say you have an entity named `comments` and you have a live view that\nshows a list of thoses comments.\n\nFor a better user experience you would like to synchronize this list in realtime\nas soon as a new comment is created.\n\nFor this matter you will use the `EventStreamex.EventListener` module in your live view:\n\n```elixir\ndefmodule MyApp.CommentLive.Index do\n  use MyApp, :live_view\n\n  use EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [%{scopes: [post_id: \"posts\"]}]\n\n  alias MyApp.Blog\n  alias MyApp.Blog.Comment\n\n  @impl true\n  def mount(params, session, socket) do\n    super(params, session, socket)\n  end\n\n  @impl true\n  def handle_params(%{\"post_id\" => post_id} = params, url, socket) do\n    {_res, socket} = super(params, url, socket)\n\n    {:noreply,\n    socket\n    |> stream(:comments, Blog.list_comments(post_id))\n    |> assign(:post_id, post_id)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_insert, [{\"posts\", post_id}], \"comments\", comment},\n        %{assigns: %{post_id: post_id}} = socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_insert(:comments, comment.new_record)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_update, [{\"posts\", post_id}], \"comments\", comment},\n        %{assigns: %{post_id: post_id}} = socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_insert(:comments, comment.new_record)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_delete, [{\"posts\", post_id}], \"comments\", comment},\n        %{assigns: %{post_id: post_id}} = socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_delete(:comments, comment.old_record)\n  end\nend\n```\n\nThis code will update the comments list as soon as a new comment is either\ninserted, deleted or udpated (notice that here, comments are linked to a `post` via the `post_id` field and we load only the comments for a specific post).\n\n**For this code to work you will also have to use the `EventStreamex.Events` module to mark the entity for WAL listening. Without doing so, the `comments` entity will not be listened from WAL events, and thus, will not be dispatched to operators and live views**","ref":"EventStreamex.EventListener.html"},{"type":"module","title":"How it works - EventStreamex.EventListener","doc":"`use EventStreamex.EventListener` will do the \"magic\" by subscribing to the entity\nchanges in `Phoenix.LiveView.mount/3`, `Phoenix.LiveView.terminate/3` and `Phoenix.LiveView.handle_params/3`\ncallbacks.\n\nThat means that is you override these callbacks you have to call the `super` function\nso that the \"magic\" is done.\n\nThe \"magic\" in question is a subscription to several channels in a pubsub module.\nThere are 3 kinds a channels this module will automatically subscribe to:\n* `direct`: We subscribe to a specific entity changes (by it's ID)\n* `unscoped`: We subscribes to changes of all entities in a table\n* `scopes`: We subscribe to changes in entities that match a specific scope (like having a specific `post_id` in the example above)\n\nWe can use the 3 kinds a subscriptions at the same time:\n\n```elixir\nuse EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [:direct, :unscoped, %{scopes: [post_id: \"posts\"]}]\n```\n\nBy default, we automatically subscribe to the `direct` and `unscoped` channels.\n\nAll events are received in `Phoenix.LiveView.handle_info/2` callbacks with a message of this form:\n\n```elixir\n{:on_insert | :on_delete | :on_update, :direct | [] | [{binary(), id()}], binary(), entity_change()}\n```\n\n*More information about each kind of message in the subsections below*\n\n### `:direct`\n\nThe `:direct` channel subscribes to a specific entity by its `id` field.\nThis is most usefull in `show` or `edit` views where you only need to show one specific entity.\n\n```elixir\nuse EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [:direct]\n```\n\nThat means that if you want to subscription to be effective, you will have to receive the id of entity\nin the params.\nOr, pass it manually when you call the `super` function:\n\n```elixir\n@impl true\ndef mount(_params, session, socket) do\n  super(%{\"id\" => get_resource_id(session)}, session, socket)\nend\n```\n\nThe changes in the entity will be received in the `Phoenix.LiveView.handle_info/2` callback\nwith a message of this form:\n\n```elixir\n{:on_insert | :on_update | :on_delete, :direct, binary(), entity_change()}\n```\n\nHere is an example:\n\n```elixir\ndefmodule MyApp.CommentLive.Show do\n  use MyApp, :live_view\n\n  use EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [:direct]\n\n  alias MyApp.Blog\n\n  @impl true\n  def mount(params, session, socket) do\n    super(params, session, socket)\n  end\n\n  @impl true\n  def handle_params(%{\"id\" => id} = params, url, socket) do\n    {_res, socket} = super(params, url, socket)\n\n    {:noreply,\n    socket\n    |> assign(:id, id)\n    |> assign(:comment, Blog.get_comment!(id))}\n  end\n\n  @impl true\n  def handle_info({:on_insert, :direct, \"comments\", _comment}, socket) do\n    # Should never happen because the entity already exists\n    {:noreply, socket}\n  end\n\n  @impl true\n  def handle_info({:on_update, :direct, \"comments\", comment}, socket) do\n    {:noreply,\n    socket\n    |> assign(:comment, comment.new_record)\n  end\n\n  @impl true\n  def handle_info({:on_delete, :direct, \"comments\", _comment}, socket) do\n    # Do some redirection stuff eventually\n    {:noreply, socket |> put_flash(:warning, \"This comment has been deleted\")}\n  end\nend\n```\n\n### `:unscoped`\n\nThe `:unscoped` channel subscribes to changes of all entities.\nThis is what you will use for `index` views when your entity is not scoped.\n\n```elixir\nuse EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [:unscoped]\n```\n\nThe changes in the entities will be received in the `Phoenix.LiveView.handle_info/2` callback\nwith a message of this form:\n\n```elixir\n{:on_insert | :on_update | :on_delete, [], binary(), entity_change()}\n```\n\nHere is an example:\n\n```elixir\ndefmodule MyApp.CommentLive.Index do\n  use MyApp, :live_view\n\n  use EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [:unscoped]\n\n  alias MyApp.Blog\n  alias MyApp.Blog.Comment\n\n  @impl true\n  def mount(params, session, socket) do\n    super(params, session, socket)\n  end\n\n  @impl true\n  def handle_params(params, url, socket) do\n    {_res, socket} = super(params, url, socket)\n\n    {:noreply,\n    socket\n    |> stream(:comments, Blog.list_comments())\n  end\n\n  @impl true\n  def handle_info(\n        {:on_insert, [], \"comments\", comment}, socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_insert(:comments, comment.new_record)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_update, [], \"comments\", comment}, socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_insert(:comments, comment.new_record)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_delete, [],\"comments\",  comment}, socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_delete(:comments, comment.old_record)\n  end\nend\n```\n\n### `:scopes`\n\nThe `:scopes` channel subscribes to changes of all entities that match some\nspecific field values.\n\nTo declare a scoped entity you will do like this:\n\n```elixir\nuse EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [%{scopes: [post_id: \"posts\"]}]\n```\n\nHere, `comments` have a `post_id` field related to a `posts` entity.\n\nYou can also have several scopes for an entity (**The order matters for the matching in `Phoenix.LiveView.handle_info/2`**):\n\n```elixir\nuse EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [%{scopes: [org_id: \"organizations\", post_id: \"posts\"]}]\n```\n\nHere, `comments` have a `org_id` field related to a `organizations` entity, and a `post_id` field, related to a `posts` entity.\n\nAs for `:direct` scopes, you will have to receive the scope fields in the params so that the module\nis able to subscribe to the correct channels.\nIf your fields are not in the params or are named differently, you will have to pass them yourself to the `super` function:\n\n```elixir\n@impl true\ndef handle_params(%{\"my_post_id\" => my_post_id, \"my_org_id\" => my_org_id} = params, url, socket) do\n  {_res, socket} = super(%{\"post_id\" => my_post_id, \"org_id\" => my_org_id}, url, socket)\n\n  {:noreply,\n  socket\n  |> stream(:comments, Blog.list_comments(my_post_id, my_org_id))\nend\n```\n\nEvents will be received in the `c:Phoenix.LiveView.handle_info/2` callback, with messages of this form:\n\n```elixir\n{:on_insert | :on_update | :on_delete, [{\"related_scoped_entity\", scope_id}], binary(), entity_change()}\n```\n\nFor instance, with our previous example, an insert event message will look like this:\n\n```elixir\n{:on_insert, [{\"organizations\", org_id}, {\"posts\", post_id}], \"comments\", entity_change}\n```\n\n**The order of scopes will be the same as the one you specified above in the `use`**.\n\nHere is a full example:\n\n\n```elixir\ndefmodule MyApp.CommentLive.Index do\n  use MyApp, :live_view\n\n  use EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [%{scopes: [post_id: \"posts\"]}]\n\n  alias MyApp.Blog\n  alias MyApp.Blog.Comment\n\n  @impl true\n  def mount(params, session, socket) do\n    super(params, session, socket)\n  end\n\n  @impl true\n  def handle_params(%{\"post_id\" => post_id} = params, url, socket) do\n    {_res, socket} = super(params, url, socket)\n\n    {:noreply,\n    socket\n    |> stream(:comments, Blog.list_comments(post_id))\n    |> assign(:post_id, post_id)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_insert, [{\"posts\", post_id}], \"comments\", comment},\n        %{assigns: %{post_id: post_id}} = socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_insert(:comments, comment.new_record)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_update, [{\"posts\", post_id}], \"comments\", comment},\n        %{assigns: %{post_id: post_id}} = socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_insert(:comments, comment.new_record)\n  end\n\n  @impl true\n  def handle_info(\n        {:on_delete, [{\"posts\", post_id}], \"comments\", comment},\n        %{assigns: %{post_id: post_id}} = socket\n      ) do\n    {:noreply,\n    socket\n    |> stream_delete(:comments, comment.old_record)\n  end\nend\n```","ref":"EventStreamex.EventListener.html#module-how-it-works"},{"type":"module","title":"Entity change structure - EventStreamex.EventListener","doc":"The event received in the `Phoenix.LiveView.handle_info/2` callbacks have information about\nthe entity and its changes.\n\nHere is what the structure looks like:\n\n```elixir\n%WalEx.Event{\n  name: atom(),\n  type: :insert | :update | :delete,\n  source: %WalEx.Event.Source{\n    name: String.t(),\n    version: String.t(),\n    db: String.t(),\n    schema: String.t(),\n    table: String.t(),\n    columns: map()\n  },\n  new_record: map() | nil,\n  old_record: map() | nil,\n  changes: map() | nil,\n  timestamp: DateTime.t(),\n  lsn: {integer(), integer()}\n}\n```\n\n*I am using the [WalEx package internally](https://github.com/cpursley/walex)*\n\n* `name`: The name of the entity (ie: `:comments` for a table named `comments`)\n* `type`: The type of event between `insert`, `update` and `delete`\n* `source`: Information about the event:\n  * `name`: \"WalEx\"\n  * `version`: Current version of `WalEx`\n  * `db`: The name of the database\n  * `schema`: Mostly `\"public\"`\n  * `table`: The name of the table (ie: `\"comments\"`)\n  * `columns`: A map of fields with their type (ie: `%{\"id\": \"integer\", \"message\": \"varchar\"}`)\n* `new_record`: The entity itself for `insert` and `update` events. `nil` for `delete` events.\n* `old_record`: The entity itself for `delete` events. `nil` for `insert` and `update` events.\n* `changes`: A map with the changes in the entity in `update` events, `nil` otherwise (see below)\n* `timestamp`: The timstamp of the event in `DateTime` type\n* `lsn`: A tuple containing information about the publication cursor\n\n### `changes`\n\nWhen you receive an `update` event, you will also have the `changes` field set to a map containing\nthe changes the entity received since the update.\n\nThis map contains the changed fields as keys, and a map describing the change as value.\nThis \"change\" map contains 2 fields:\n* `old_value`: The value before the update\n* `new_value`: The value after the update\n\nFor instance, let's say you have a `comments` entity with 4 fields: `id`, `message`, `rating`, `timestamp`.\n\nYou have a comment with these values:\n\n```elixir\n%Comment{\n  id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n  message: \"This is my first comment\",\n  rating: 4,\n  timestamp: \"2024-07-23T12:00:00Z\"\n}\n```\n\nNow, the comment is update this these new values:\n\n```elixir\n%Comment{\n  id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n  message: \"This is (not) my first comment (anymore)\",\n  rating: 5,\n  timestamp: \"2024-07-23T12:00:00Z\"\n}\n```\n\nThe event structure will look like this:\n\n```elixir\n%WalEx.Event{\n  name: :comments,\n  type: :update,\n  source: %WalEx.Event.Source{\n    name: \"WalEx\",\n    version: \"4.4.0\",\n    db: \"postgresql\",\n    schema: \"public\",\n    table: \"comments\",\n    columns: %{\n      id: \"uuid\",\n      message: \"varchar\",\n      rating: \"integer\",\n      timestamp: \"datetime\"\n    }\n  },\n  new_record: %Comment{\n    id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n    message: \"This is (not) my first comment (anymore)\",\n    rating: 5,\n    timestamp: \"2024-07-23T12:00:00Z\"\n  },\n  old_record: nil,\n  changes: %{\n    message: %{\n      old_value: \"This is my first comment\",\n      new_value: \"This is (not) my first comment (anymore)\"\n    },\n    rating: %{\n      old_value: 4,\n      new_value: 5\n    }\n  },\n  timestamp: \"2024-08-25T13:13:30Z\",\n  lsn: {0, 0}\n}\n```","ref":"EventStreamex.EventListener.html#module-entity-change-structure"},{"type":"module","title":"Unsubscribing from events - EventStreamex.EventListener","doc":"The unsubscribe from events is done automatically in the `c:Phoenix.LiveView.terminate/3` callback.\nYou do not have anything to do except for calling the `super` function if you override this callback.","ref":"EventStreamex.EventListener.html#module-unsubscribing-from-events"},{"type":"module","title":"Handling subscriptions later - EventStreamex.EventListener","doc":"If you need, for any reason, to handle subscriptions at another moment than the `mount` and `handle_params` callbacks,\nwe provide the `handle_subscriptions/2` function.\n\nThis can be useful if the parameters used for scoped channels\nare handled in a different way than just getting them from query parameters.","ref":"EventStreamex.EventListener.html#module-handling-subscriptions-later"},{"type":"module","title":"Params - EventStreamex.EventListener","doc":"* `socket`: The current socket\n* `params`: A map containing the parameters to use for scoped channels. Each scope field must be present in the params map as a string","ref":"EventStreamex.EventListener.html#module-params"},{"type":"module","title":"Returns - EventStreamex.EventListener","doc":"The updated socket","ref":"EventStreamex.EventListener.html#module-returns"},{"type":"module","title":"Example - EventStreamex.EventListener","doc":"```elixir\ndefmodule MyModule do\n  use EventStreamex.EventListener,\n    schema: \"comments\",\n    subscriptions: [%{scopes: [post_id: \"posts\"]}]\n\n  def handle_params(%{\"id\" => id}, _url, socket) do\n    entity = MyEntity.get(id)\n\n    {:noreply,\n      handle_subscriptions(\n        socket,\n        %{\"post_id\" => entity.post_id}\n      )}\n  end\nend\n```\n\nIn this example, I need a scoped channel with a `post_id` field.\nThis field is not present in the url (which is a bad practice), but\nI know I can find it in the entity I fetch.\n\nHere, I don't have to call the `super()` function because I do not need\nthe `EventListener` to handle the subscriptions as I do it manually.\n\nYou don't have to handle the unsubscribe either because it will be done for you in the\n`Phoenix.LiveView.terminate/3` callback.","ref":"EventStreamex.EventListener.html#module-example"},{"type":"module","title":"Subscribing to other entities' events - EventStreamex.EventListener","doc":"Let's say you display posts and handle their creation.\nAnd when a new post is created you want to redirect to the detail page\nof this new post.\nBut, the enitty used for the detail of this post is a derived entity from\nan operator.\nThus, you can't be sure the entity will exist when you redirect to its detail page\nbecause of this asynchronicity of the event streaming architecture.\n\nSo you would like to listen for this entity beeing created.\nHopefully, you know its ID because you just created the post and this derived\nentity uses the same ID.\n\nSo you can call the `subscribe_entiy/4` function to listen for the events coming from\nthis derived entity.\n\n```elixir\nsubscribe_entity(socket, \"post_with_comments_count\", :direct, %{\"id\" => post_id})\n```","ref":"EventStreamex.EventListener.html#module-subscribing-to-other-entities-events"},{"type":"module","title":"Params - EventStreamex.EventListener","doc":"* `socket`: The socket\n* `entity_name`: The entity name to listen to as a string\n* `subscription`: The kind of channel you want to listen to (these are the same as for the module configuration: `:direct`, `:unscoped`, `%{scopes: []}`)\n* `params`: A map with the parameters needed for `:direct` and `:scopes` scopes.","ref":"EventStreamex.EventListener.html#module-params"},{"type":"module","title":"Return value - EventStreamex.EventListener","doc":"The updated `socket`","ref":"EventStreamex.EventListener.html#module-return-value"},{"type":"module","title":"Received events - EventStreamex.EventListener","doc":"Events are received the same way as other events but related to the entity:\n\n```elixir\n@impl true\ndef handle_info({:on_insert, :direct, \"post_with_comments_count\", post}, socket) do\n  {:noreply, socket |> push_navigate(to: \"/posts/#{post.id}\")}\nend\n```","ref":"EventStreamex.EventListener.html#module-received-events"},{"type":"module","title":"Unsubscribing to other entities' events - EventStreamex.EventListener","doc":"Use the function `unsubscribe_entity/3` to manually unsubscribe from an event's events.\n\n```elixir\nunsubscribe_entity(socket, \"post_with_comments_count\", :direct)\n```\n\nThe params will be the same used for subscription so no need to pass them again.","ref":"EventStreamex.EventListener.html#module-unsubscribing-to-other-entities-events"},{"type":"module","title":"Params - EventStreamex.EventListener","doc":"* `socket`: The socket\n* `entity_name`: The name of the entity to stop linstening\n* `subscription`: The kind of channel to stop listening to (these are the same as for the module configuration: `:direct`, `:unscoped`, `%{scopes: []}`)","ref":"EventStreamex.EventListener.html#module-params"},{"type":"module","title":"Return value - EventStreamex.EventListener","doc":"The updated socket\n\n## `use` params\n\nWhen you `use` this module, you will be able to specify these parameters:\n\n* `:schema`: The name of the entity as a string (mandatory field)\n* `:subscriptions`: The subscriptions you want to do (Defaults to `[:direct, :unscoped]`)\n* `:application`: The application module to use to retrieve config values (Defaults to `Application`)","ref":"EventStreamex.EventListener.html#module-return-value"},{"type":"module","title":"EventStreamex.Events","doc":"This module can be used in schema modules and is necessary to mark an\nentity for event listening.\n\nThis is what connects an entity to the WAL events.\n\nIf you don't use this module for an entity you won't be able to receive\nits events, which will prevent you from listening to events in live views (`EventStreamex.EventListener`),\nand to use them in operators to update other entities (`EventStreamex.Operators.Operator`).\n\nTo mark an entity for event listening, you only have to `use` this module in\nthe schema module of the entity:\n\n```elixir\ndefmodule MyApp.Blog.Comment do\n  use MyApp.Schema\n\n  use EventStreamex.Events,\n    name: :my_app,\n    schema: \"comments\",\n    extra_channels: [%{scopes: [post_id: \"posts\"]}]\n\n  import Ecto.Changeset\n\n  schema \"comments\" do\n    field :message, :string\n    field :post_id, Ecto.UUID\n\n    timestamps(type: :utc_datetime)\n  end\n\n  @doc false\n  def changeset(comment, attrs) do\n    comment\n    |> cast(attrs, [:id, :message, :post_id])\n    |> validate_required([:message, :post_id])\n  end\nend\n```\n\nIn this example, we have a `comments` entity that we want to mark for event listening.\n\nThis `comments` entity is scoped by a `post_id` related to a `posts` entity.\nSo we needed to add an `extra_channels` so that comment events will also be emitted in a channel\nspecific for his scopes (See `EventStreamex.EventListener` for more details on scopes)\n\nSeveral scopes can be used for an entity. The order used for scopes must be the same\nwhen you define it in `EventStreamex.EventListener`.\n\nHere is an example with another scope:\n\n```elixir\nuse EventStreamex.Events,\n    name: :my_app,\n    schema: \"comments\",\n    extra_channels: [%{scopes: [org_id: \"organizations\", post_id: \"posts\"]}]\n```\n\nHere we say a comment has 2 scopes: an organization and a post.\nThat means the entity must have the 2 fields: `org_id` and `post_id`.","ref":"EventStreamex.Events.html"},{"type":"module","title":"How it works - EventStreamex.Events","doc":"As said earlier, this module marks the entity for WAL events.\nWhich allows you to listen to its changes and to use it in operators.\n\nUnder the hoods, this module will create a sub module inside the module you use it in.\nAnd it will define functions that will be called each time something happens with this entity (insert, update, delete).\n\nThen, it will dispatch the event to all operators listening to this entity (`EventStreamex.Operators.Operator`).\nAnd it will emit an event in the configured pubsub module in several channels:\n\n* `direct`: We emit in a specific channel containing the ID of the enity\n* `unscoped`: We emit in a channel for changes of all entities of this type\n* `scopes`: We emit in a channel for entities matching the given scopes (Like in the example above)\n\n*For more information about scopes and channels, see `EventStreamex.EventListener`.*\n\n## `use` params\n\nWhen you `use` this module, you will be able to specify these parameters:\n\n* `:module_name`: The name of the module that will be created in your module to perform the events logic (Defaults to `\"Events\"`)\n* `:schema`: The name of the entity as a string (mandatory field)\n* `:name`: The name of your application (Mandatory field)\n* `:extra_channels`: The channels to emit events to. `:direct` and `:unscoped` are always used, and you can add your own scopes for this entity (Defaults to `[]`)\n* `:application`: The application module to use to retrieve config values (Defaults to `Application`)\n* `:scheduler`: If you want to use another scheduler to handle events processed by operators (Defaults to `EventStreamex.Operators.Scheduler`)","ref":"EventStreamex.Events.html#module-how-it-works"},{"type":"function","title":"EventStreamex.Events.schema_module/1","doc":"","ref":"EventStreamex.Events.html#schema_module/1"},{"type":"module","title":"EventStreamex.Operators.EntityProcessed","doc":"This module is called last for each event.\n\nAll it does is register the last updated timestamp of the entity and store it\nin the `EventStreamex.Operators.ProcessStatus` module.\n\nBy default the timestamp is taken from the `updated_at` field of the entity\nbut you can change that in the `:process_status_entity_field` config param.\n\nIf no timestamp could be found, the entity is not registered in the process status.\n\nWorks with fields of type:\n* `bigint`: The unix timestamp with a microsecond precision\n* `DateTime`: Using microsecond precision\n* `NaiveDateTime`: Using microsecond precision","ref":"EventStreamex.Operators.EntityProcessed.html"},{"type":"function","title":"EventStreamex.Operators.EntityProcessed.processed/1","doc":"","ref":"EventStreamex.Operators.EntityProcessed.html#processed/1"},{"type":"function","title":"EventStreamex.Operators.EntityProcessed.to_unix_timestamp/1","doc":"Transforms the last updated time field value to a unix timestamp.\nThe precision must be in microsecond.\n\nAllowed types are:\n* `bigint`: The unix timestamp with a microsecond precision\n* `DateTime`: Using microsecond precision\n* `NaiveDateTime`: Using microsecond precision\n\nReturns `{:ok, unix_timestamp}` or `{:error, reason}`","ref":"EventStreamex.Operators.EntityProcessed.html#to_unix_timestamp/1"},{"type":"function","title":"Examples - EventStreamex.Operators.EntityProcessed.to_unix_timestamp/1","doc":"```elixir\niex> EventStreamex.Operators.EntityProcessed.to_unix_timestamp(nil)\n{:error, :no_value}\niex> EventStreamex.Operators.EntityProcessed.to_unix_timestamp(\"bad arg\")\n{:error, :bad_arg}\niex> EventStreamex.Operators.EntityProcessed.to_unix_timestamp(1422057007123000)\n{:ok, 1422057007123000}\niex> {:ok, datetime, 9000} = DateTime.from_iso8601(\"2015-01-23T23:50:07.123+02:30\")\niex> EventStreamex.Operators.EntityProcessed.to_unix_timestamp(datetime)\n{:ok, 1422048007123000}\niex> naivedatetime = NaiveDateTime.from_iso8601!(\"2015-01-23T23:50:07.123Z\")\niex> EventStreamex.Operators.EntityProcessed.to_unix_timestamp(naivedatetime)\n{:ok, 1422057007123000}\n```","ref":"EventStreamex.Operators.EntityProcessed.html#to_unix_timestamp/1-examples"},{"type":"function","title":"EventStreamex.Operators.EntityProcessed.to_unix_timestamp!/1","doc":"Transforms the last updated time field value to a unix timestamp.\n\nSee `to_unix_timestamp/1`","ref":"EventStreamex.Operators.EntityProcessed.html#to_unix_timestamp!/1"},{"type":"module","title":"EventStreamex.Operators.Executor","doc":"Executes the operator tasks when an event is received.\n\nA new executor is started by the `EventStreamex.Operators.Scheduler` as soon\nas an event is received.\nIt is responsible to start the operator the scheduler gave it to handle the\nnew event.\nAnd it will handle the operator's failures.\n\nMeaning that if the operator crashes or returns an error, the executor will\nrestart until the operator successfully finishes its task or has failed to many times.","ref":"EventStreamex.Operators.Executor.html"},{"type":"module","title":"Restart strategy - EventStreamex.Operators.Executor","doc":"In the configuration you can set the restart strategy:\n\n```elixir\nconfig :event_streamex,\n  operator_queue_backoff_multiplicator: 2,\n  operator_queue_max_restart_time: 10000,\n  operator_queue_max_retries: 5,\n  operator_queue_min_restart_time: 500\n```\n\n* `operator_queue_backoff_multiplicator`: At each new retry, the last wait time is multiplied by this multiplicator to get the new time (ie: If it waited 1.5 second for the last retry and the multiplicator is 2, then, it will now wait 3 seconds before the next restart).\n* `operator_queue_max_restart_time`: The time we wait before a restart is maxed at this value.\n* `operator_queue_max_retries`: The maximum number of times we allow the operators to retry after a failure.\n* `operator_queue_min_restart_time`: The minimum time we will wait before a restart.\n\nAt each failure of an operator, the error logger is called (`EventStreamex.Operators.Logger.ErrorLoggerAdapter`)\nas well as a telemetry.\n\nWhen an operator exceeds the maximum number of restarts the executor crashes as\nwell as the `EventStreamex.Operators.Scheduler` and no more event will be processed until you\nrestart the scheduler with `EventStreamex.restart_scheduler/0`.\n\nThis is because the order of the events is very important and we cannot allow an\nevent to fail, otherwise it would compromise your data.\n\nYou can create your own `EventStreamex.Operators.Logger.ErrorLoggerAdapter` and add it to the configuration:\n\n```elixir\nconfig :event_streamex,\n  error_logger_adapter: {EventStreamex.Operators.Logger.LoggerAdapter, []}\n```\n\nDoing so, you could implement some kind of notification system when there is a\nfailure to warn your team about it.\n\nWhen a failure happens, maybe it's because of a temporary failure and you could just restart the scheduler.\nBut maybe the failure is due to a coding failure.\nIn this case you should correct the issue, ship a new version of your app and restart the scheduler (or the entire application).\n\nWith a database queue adapter (the default one), the queue is saved to database and will be loaded again when the\napplication restarts.\nYou won't miss any event.\n\n**The executor is used internally, you should never call it directly**","ref":"EventStreamex.Operators.Executor.html#module-restart-strategy"},{"type":"function","title":"EventStreamex.Operators.Executor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Executor.html#child_spec/1"},{"type":"behaviour","title":"EventStreamex.Operators.Logger.ErrorLoggerAdapter","doc":"Behaviour to handle errors with operators (`EventStreamex.Operators.Operator`).\n\nWhen implementing this behaviour, you must put use a `GenServer`.\n\nThere is currently one adapter available and used by default: `EventStreamex.Operators.Logger.LoggerAdapter`.\nThis adapter simply logs the error to the console.\n\nThe executor (`EventStreamex.Operators.Executor`) who catches the operator failure and\ncalls the error logger will also execute a telemetry event so that you can\nsee errors over time in a chart.\n\nThere are 2 kinds of errors:\n* **Retry**: The operator failed but will retry soon\n* **Failed**: The operator retried several times but failed again.\n\nFor more information about the retry strategy, see `EventStreamex.Operators.Executor`.","ref":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.html"},{"type":"function","title":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.html#child_spec/1"},{"type":"callback","title":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.log_failed/3","doc":"Same as `log_retry/3` but when no more retry will be attempted\n\n# Params\n\n* `module`: The module of the operator\n* `reason`: The failure reason\n* `retry_status`: The current retry count, the time to wait for the next retry, etc...","ref":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.html#c:log_failed/3"},{"type":"callback","title":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.log_retry/3","doc":"Logs an error with the operator, that will be retried again soon.\n\n# Params\n\n* `module`: The module of the operator\n* `reason`: The failure reason\n* `retry_status`: The current retry count, the time to wait for the next retry, etc...","ref":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.html#c:log_retry/3"},{"type":"callback","title":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.start_link/1","doc":"This function will be called at startup to spawn the error logger adapter.\n\nThe param comes from the configuration of the adapter:\n\n```elixir\nerror_logger_adapter: {EventStreamex.Operators.Logger.LoggerAdapter, []},\n```\n\nThe return value is a tuple with the pid of the spawned process","ref":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.html#c:start_link/1"},{"type":"type","title":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.retry_status/0","doc":"The retry status:\n\n* `max_retries`: Maximum number of retries attempted\n* `curr_retries`: The current retries count already attempted (starting at 1 after the first failure)\n* `max_restart_time`: The maximum number of milliseconds to wait for the next retry\n* `backoff_multiplicator`: A decimal number multiplied to the `curr_time_to_wait` to define the next amount of time to wait for the next retry (maxed by `max_restart_time`)\n* `curr_time_to_wait`: The current time to wait in milliseconds before the next retry","ref":"EventStreamex.Operators.Logger.ErrorLoggerAdapter.html#t:retry_status/0"},{"type":"module","title":"EventStreamex.Operators.Logger.LoggerAdapter","doc":"An error logger adapter that writes logs in the console.\n\nSee `EventStreamex.Operators.Logger.ErrorLoggerAdapter` for more information\nabout the error logging.","ref":"EventStreamex.Operators.Logger.LoggerAdapter.html"},{"type":"function","title":"EventStreamex.Operators.Logger.LoggerAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Logger.LoggerAdapter.html#child_spec/1"},{"type":"function","title":"EventStreamex.Operators.Logger.LoggerAdapter.log_failed/3","doc":"Logs an error with the `:critical` level.\nThe operator will not be restarted again.","ref":"EventStreamex.Operators.Logger.LoggerAdapter.html#log_failed/3"},{"type":"function","title":"EventStreamex.Operators.Logger.LoggerAdapter.log_retry/3","doc":"Logs an error with the `:error` level.\nThe operator will be restarted.","ref":"EventStreamex.Operators.Logger.LoggerAdapter.html#log_retry/3"},{"type":"function","title":"EventStreamex.Operators.Logger.LoggerAdapter.start_link/1","doc":"Starts the adapter","ref":"EventStreamex.Operators.Logger.LoggerAdapter.html#start_link/1"},{"type":"module","title":"EventStreamex.Operators.Operator","doc":"An operator listens for events and will usually create/update/delete\nanother entity based on those events.\n\nThis is a very usefull pattern. Here is why:","ref":"EventStreamex.Operators.Operator.html"},{"type":"module","title":"The issue without event streaming - EventStreamex.Operators.Operator","doc":"Let's say you have a `posts` entity which contains:\n* `title`: a title\n* `body`: a body\n\nYou also have a `comments` entity which contains:\n* `post_id`: The ID of the post it is attached to\n* `comment`: The comment content\n\nWhen you show a list of your posts you might want to show the number\nof comments it contains.\n\nTo do so you will be tempted to make a SQL query with a subquery that will\ncount the comments of your posts.\n\nThis is pretty trivial:\n\n```sql\nSELECT\n  p.id AS id,\n  p.title AS title,\n  p.body AS body,\n  (SELECT COUNT(*) FROM comments c WHERE c.post_id = p.id) AS comments_count\nFROM posts p;\n```\n\nBut it has several flaws.\nLet's see them in more details.","ref":"EventStreamex.Operators.Operator.html#module-the-issue-without-event-streaming"},{"type":"module","title":"Your queries will become slower and slower over time - EventStreamex.Operators.Operator","doc":"Because subqueries are not free. They take more time to execute that a simple\nquery with no join and subquery.\n\nSo, the more posts and comments you have, the slower the query will be.","ref":"EventStreamex.Operators.Operator.html#module-your-queries-will-become-slower-and-slower-over-time"},{"type":"module","title":"You will end up adding complexity in your queries - EventStreamex.Operators.Operator","doc":"As your app evolves, you will probably create new entities related to your posts and comments.\nThus, you will add more and more subqueries and joins to your queries, making them slower and harder to maintain...","ref":"EventStreamex.Operators.Operator.html#module-you-will-end-up-adding-complexity-in-your-queries"},{"type":"module","title":"Realtime will become a mess to maintain - EventStreamex.Operators.Operator","doc":"If you want to handle realtime events like updating the number of comments when a new one\nis created it will be a mess to maintain overtime because you will end up subscribing to a lot of events\nas you link your posts to other entities.\n\nAnd you will probably forget to subscribe to some of the events making your app\ninconsistant (some features will be in realtime and other won't be).","ref":"EventStreamex.Operators.Operator.html#module-realtime-will-become-a-mess-to-maintain"},{"type":"module","title":"Almost impossible to maintain recursivity - EventStreamex.Operators.Operator","doc":"If your comments can have subcomments it will be almost impossible\nto maintain the comments count because of the recursivity it implies.","ref":"EventStreamex.Operators.Operator.html#module-almost-impossible-to-maintain-recursivity"},{"type":"module","title":"Depending on entities not directly related to the base entity - EventStreamex.Operators.Operator","doc":"Let's say you create another entity named `ratings` which saves each rating given\nto a specific comment.\n\nIt would look like this:\n* `comment_id`: The ID of the comment related to this rating\n* `rating`: The given rating\n\nNow imagine you want to display for every comment, the average rating and\nthe total count of ratings for this comment.\n\nHere again, you could solve this with some subqueries.\n\nBut what if you want to show the average rating of all comments in the post\nso that users can see that a post has received a lot of useful comments.\n\nI don't even want to think about the query to build to do that...","ref":"EventStreamex.Operators.Operator.html#module-depending-on-entities-not-directly-related-to-the-base-entity"},{"type":"module","title":"Benefits from using event streaming - EventStreamex.Operators.Operator","doc":"Let's come back to our `posts` and `comments`.\nNow let's imagine that instead of our subquery to count the comments we use\nan event streaming architecture and create another entity whose goal is\nspecifically to show the posts with their comments count.\n\nThis `post_with_comments_counts` entity will look like that:\n* `title`: a title\n* `body`: a body\n* `comments_count`: the number of comments of the post\n\nAnd this entity will listen to new comments to update its comments count:\nwhen a new comment is created it will increase the number of comments of the related post by 1.\n\nWith `EventStreamex` it would look like that:\n\n```elixir\ndefmodule MyApp.Operators.PostWithCommentsCountOperator do\n  use EventStreamex.Operators.Operator,\n    schema: \"post_with_comments_counts\",\n    repo: MyApp.Repo\n\n  on_event(\n    all_events(),\n    entities([\"comments\", \"posts\"]),\n    query(\n      \"MERGE INTO post_with_comments_counts dest\n      USING (\n          SELECT\n              p.id AS id,\n              p.title AS title,\n              p.body AS body,\n              p.inserted_at AS inserted_at,\n              p.updated_at AS updated_at,\n              (SELECT COUNT(c.id) FROM comments c WHERE c.post_id = p.id) AS comments_count\n          FROM posts p\n          WHERE p.id = $1\n      ) src ON src.id = dest.id\n      WHEN NOT MATCHED THEN\n        INSERT (id, title, body, inserted_at, updated_at, comments_count)\n        VALUES (src.id, src.title, src.body, src.inserted_at, src.updated_at, src.comments_count)\n      WHEN MATCHED THEN\n        UPDATE SET\n            title = src.title,\n            body = src.body,\n            inserted_at = src.inserted_at,\n            updated_at = src.updated_at,\n            comments_count = src.comments_count;\",\n      id([{\"comments\", :post_id}, {\"posts\", :id}]),\n      delete_with([\"posts\"])\n    )\n  )\nend\n```\n\nHere, we use the `EventStreamex.Operators.Operator` module to listen for\n`posts` and `comments` in all kinds of events (insert, update, delete).\n\nAnd each time a post or a comment is inserted, updated, deleted, we perform this `MERGE` query\nto create/update our `post_with_comments_counts`.\n\nThe `MERGE` query will insert a new post if the post does not exist and update\nit if it already exists.\n\n`id/1` is used to map the fields of the listened entities to the id of the\n`post_with_comments_counts` entity (`id` for a post and `post_id` for a comment).\n\nFinally, `delete_with/1` tells the operator which entities deletion should\nalso delete the related `post_with_comments_counts` entity.\n\nAs you can see, this `MERGE` query also uses a subquery, so why is it better\nthan our first implementation?","ref":"EventStreamex.Operators.Operator.html#module-benefits-from-using-event-streaming"},{"type":"module","title":"This is a naive implementation - EventStreamex.Operators.Operator","doc":"The `query/3` macro is provided for simplicity.\nBut you can also replace it by a function to do things as you think it's best.\n\nThis function will receive the event, allowing you to do whatever you want with it.","ref":"EventStreamex.Operators.Operator.html#module-this-is-a-naive-implementation"},{"type":"module","title":"The query is executed in the background - EventStreamex.Operators.Operator","doc":"In our first example, the query was executed everytime someone wants to\ndisplay the list of posts.\n\nIn our event streaming example, the query is executed in the background\neach time there is an event with a post or a comment.\n\nTo display your list of posts you will just have to do:\n\n```sql\nSELECT * FROM post_with_comments_counts;\n```\n\nWhich is simpler and faster than our former query.","ref":"EventStreamex.Operators.Operator.html#module-the-query-is-executed-in-the-background"},{"type":"module","title":"Realtime is easy - EventStreamex.Operators.Operator","doc":"You can use the module `EventStreamex.EventListener` to easily add realtime\nin your live view applications.\n\nBecause we are using PostgreSQL WAL events, everytime something happens on an entity\n(`post_with_comments_counts` as well), an event is sent.\n\nThis event will be processed by operators listening to it (like our operator here does for `posts` and `comments`),\nand it will be sent to a pubsub processor.\n\nThe `EventStreamex.EventListener`, now just has to listen to this pubsub processor\nto notify your live view the entity your are showing as been updated.\n\nEverything, now, can be realtime, without beeing hard to maintain.","ref":"EventStreamex.Operators.Operator.html#module-realtime-is-easy"},{"type":"module","title":"A better pattern to model your data - EventStreamex.Operators.Operator","doc":"With this approach you keep your entities consistent and simple.\nAnd you create specialized entities for what you want to show in your views.\n\nA post will stay a post with its `title` and `body` fields and a comment will\nstay a comment with a `comment` field.\n\n`post_with_comments_count` is just a specialized entity used for your view.\n\nThis keeps your data and APIs consistent and simple with explicit fields and names.\n\nFor instance, let's say you have a `users` entity with a `login` and a `password` field.\nBut later on, you realize that you need to store some information about your users,\nlike its name, its address, and so on.\n\nYou will be tempted to update your `users` entity to add these new fields:\n* `login`\n* `password`\n* `address`\n* `name`\n\nBut still, in your application, when comes the moment to create your user,\nyou only need the `login` and `password`.\n\nSo what do you do? You set the other fields as optionals in your API and set them\na default value (empty string or null).\n\nAnd in the view that allows a user to specify more information about him (`name` and `address`),\nyou will either:\n* Use the same API endpoint as for the `users` creation\n* Or create a new dedicated API endpoint\n\nBut both approachs are bad practices!\n\nIn the first approach that means that your API expects 2 mandatory fields: `login` and `password`\nand 2 optional fields: `name` and `address`.\nBut here, there is no sense in sending the `login` and `password`.\nAnd maybe you would like the `name` and `address` fields to be mandatory.\nBut because your API set these fields as optionals, you will have to check for them from the client side.\nWhich is a bad practice.\n\nIn the second approach, that means that you will create an API endpoint for the user profile.\nBut this API, will finally update the `users` entity.\nYou end up with hard to understand APIs because their name does not match\nto the entity being modified.\n\nA better approach would be to have a `users` entity with:\n* `login`\n* `password`\nAnd its own API endpoint `/users` with only these two fields.\n\nAnd you would have a `user_profiles` entity with:\n* `user_id`\n* `name`\n* `address`\nAnd its own API endpoint `/users/{user_id}/user_profiles` with only the `name` and `address` fields.\n\nAnd what if you need to display all of these information in your view?\nWell, you use an operator like we did for the `post_with_comments_counts`.\nNamed, for instance: `user_with_profiles`, and listening to `users` and `user_profiles` events.\n\nThis way, you keep your data and API consistent.\n\n## `use` params\n\nWhen you `use` this module, you will be able to specify these parameters:\n\n* `:schema`: The name of the entity as a string (mandatory field)\n* `:repo`: The repository used for the connection to the database (mandatory field)\n* `:sql_adapter`: The SQL adapter to use (Defaults to `Ecto.Adapters.SQL`)\n* `:id`: The field used as an ID for the `schema` (Defaults to `:id`)","ref":"EventStreamex.Operators.Operator.html#module-a-better-pattern-to-model-your-data"},{"type":"macro","title":"EventStreamex.Operators.Operator.all_events/0","doc":"Macro used as first parameter to `on_event/3` to listen to all kinds of events:\n* `:on_insert`\n* `:on_update`\n* `:on_delete`","ref":"EventStreamex.Operators.Operator.html#all_events/0"},{"type":"macro","title":"EventStreamex.Operators.Operator.delete_with/1","doc":"Sets the entities that must delete the final entity when deleted.\n\nIt takes a list of entities (as a list of strings).\n\nThis macro is used as third argument to the `query/3` macro.","ref":"EventStreamex.Operators.Operator.html#delete_with/1"},{"type":"macro","title":"Example - EventStreamex.Operators.Operator.delete_with/1","doc":"```elixir\ndelete_with([\"posts\"])\n```","ref":"EventStreamex.Operators.Operator.html#delete_with/1-example"},{"type":"macro","title":"EventStreamex.Operators.Operator.entities/1","doc":"Macro used as second parameter to `on_event/3` to listen to a list of entities.\nYou have to use the name of the table in string.","ref":"EventStreamex.Operators.Operator.html#entities/1"},{"type":"macro","title":"EventStreamex.Operators.Operator.entity/1","doc":"Macro used as second parameter to `on_event/3` to listen to only one entity.\nYou have to use the name of the table in string.","ref":"EventStreamex.Operators.Operator.html#entity/1"},{"type":"macro","title":"EventStreamex.Operators.Operator.event/1","doc":"Macro used as first parameter to `on_event/3` to listen to only one event type from these items:\n* `:on_insert`\n* `:on_update`\n* `:on_delete`","ref":"EventStreamex.Operators.Operator.html#event/1"},{"type":"macro","title":"EventStreamex.Operators.Operator.events/1","doc":"Macro used as first parameter to `on_event/3` to listen to a list of event types from these items:\n* `:on_insert`\n* `:on_update`\n* `:on_delete`","ref":"EventStreamex.Operators.Operator.html#events/1"},{"type":"macro","title":"EventStreamex.Operators.Operator.id/1","doc":"Sets the mapping of ids.\n\nThe mapping is a list of tuples containing:\n* The entity name (as string)\n* The field that corresponds to the final entity's id\n\nThis macro is used as second argument to the `query/3` macro.","ref":"EventStreamex.Operators.Operator.html#id/1"},{"type":"macro","title":"Example - EventStreamex.Operators.Operator.id/1","doc":"```elixir\nid([{\"comments\", :post_id}, {\"posts\", :id}])\n```","ref":"EventStreamex.Operators.Operator.html#id/1-example"},{"type":"macro","title":"EventStreamex.Operators.Operator.no_delete/0","doc":"Specifies that no entity should delete the final entity when deleted.\n\nThis macro is used as third argument to the `query/3` macro.","ref":"EventStreamex.Operators.Operator.html#no_delete/0"},{"type":"macro","title":"EventStreamex.Operators.Operator.on_event/3","doc":"This macro is the one listening the events.\n\nYou have to define one and only one in your operator.","ref":"EventStreamex.Operators.Operator.html#on_event/3"},{"type":"macro","title":"Parameters - EventStreamex.Operators.Operator.on_event/3","doc":"* `events`: A list of event types to listen to (`:insert`, `:update`, `:delete`). You can use the `all_events/0`, `events/1` and `event/1` macros to help you.\n* `entities`: A list of entities to list to (use their table name as string). You can use `entities/1` and `entity/1` macros to help you.\n* `callback`: A function that takes the event as parameter and must return a result. You can also use the `query/3` macro instead.","ref":"EventStreamex.Operators.Operator.html#on_event/3-parameters"},{"type":"macro","title":"Example - EventStreamex.Operators.Operator.on_event/3","doc":"```elixir\non_event(\n  all_events(),\n  entities([\"comments\", \"posts\"]),\n  fn event ->\n    {:ok, res} =\n      event\n      |> MyQuery.do_some_queries()\n      |> MyQuery.then_another_query()\n    {:ok, res}\n  end\n)\n```","ref":"EventStreamex.Operators.Operator.html#on_event/3-example"},{"type":"macro","title":"EventStreamex.Operators.Operator.query/3","doc":"Macro used as third parameter to `on_event/3` to directly perform a query.","ref":"EventStreamex.Operators.Operator.html#query/3"},{"type":"macro","title":"Parameters - EventStreamex.Operators.Operator.query/3","doc":"* `query`: The query to execute for each event\n* `ids_mapping`: A list of tuples to map the correct entities' field to the id of the final entity. You can use the `id/1` macro.\n* `delete_with`: A list of entities that, when deleted, will also delete the final entity. You can use the `delete_with/1` and `no_delete/0` macros.","ref":"EventStreamex.Operators.Operator.html#query/3-parameters"},{"type":"macro","title":"Example - EventStreamex.Operators.Operator.query/3","doc":"```elixir\non_event(\n  all_events(),\n  entities([\"comments\", \"posts\"]),\n  query(\n    \"MERGE INTO post_with_comments_counts dest\n    USING (\n        SELECT\n            p.id AS id,\n            p.title AS title,\n            p.body AS body,\n            p.inserted_at AS inserted_at,\n            p.updated_at AS updated_at,\n            (SELECT COUNT(c.id) FROM comments c WHERE c.post_id = p.id) AS comments_count\n        FROM posts p\n        WHERE p.id = $1\n    ) src ON src.id = dest.id\n    WHEN NOT MATCHED THEN\n      INSERT (id, title, body, inserted_at, updated_at, comments_count)\n      VALUES (src.id, src.title, src.body, src.inserted_at, src.updated_at, src.comments_count)\n    WHEN MATCHED THEN\n      UPDATE SET\n          title = src.title,\n          body = src.body,\n          inserted_at = src.inserted_at,\n          updated_at = src.updated_at,\n          comments_count = src.comments_count;\",\n    id([{\"comments\", :post_id}, {\"posts\", :id}]),\n    delete_with([\"posts\"])\n  )\n)\n```\n\nYour query will have a `$1` parameter which will correspond to the id of the final entity.\nThis id is taken from the entity who emitted the event, based on the `ids_mapping` field.\n\nFor instance, for a `comments` event, the final entity will use the `post_id` field as his own id as set in\nthe `ids_mapping` field.\n\nThis mapping is also used in the deletion of the entity.","ref":"EventStreamex.Operators.Operator.html#query/3-example"},{"type":"type","title":"EventStreamex.Operators.Operator.event/0","doc":"The event received by the operator\n\n*I am using the [WalEx package internally](https://github.com/cpursley/walex)*\n\n* `name`: The name of the entity (ie: `:comments` for a table named `comments`)\n* `type`: The type of event between `insert`, `update` and `delete`\n* `source`: Information about the event:\n  * `name`: \"WalEx\"\n  * `version`: Current version of `WalEx`\n  * `db`: The name of the database\n  * `schema`: Mostly `\"public\"`\n  * `table`: The name of the table (ie: `\"comments\"`)\n  * `columns`: A map of fields with their type (ie: `%{\"id\": \"integer\", \"message\": \"varchar\"}`)\n* `new_record`: The entity itself for `insert` and `update` events. `nil` for `delete` events.\n* `old_record`: The entity itself for `delete` events. `nil` for `insert` and `update` events.\n* `changes`: A map with the changes in the entity in `update` events, `nil` otherwise (see below)\n* `timestamp`: The timstamp of the event in `DateTime` type\n* `lsn`: A tuple containing information about the publication cursor\n\n### `changes`\n\nWhen you receive an `update` event, you will also have the `changes` field set to a map containing\nthe changes the entity received since the update.\n\nThis map contains the changed fields as keys, and a map describing the change as value.\nThis \"change\" map contains 2 fields:\n* `old_value`: The value before the update\n* `new_value`: The value after the update\n\nFor instance, let's say you have a `comments` entity with 4 fields: `id`, `message`, `rating`, `timestamp`.\n\nYou have a comment with these values:\n\n```elixir\n%Comment{\n  id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n  message: \"This is my first comment\",\n  rating: 4,\n  timestamp: \"2024-07-23T12:00:00Z\"\n}\n```\n\nNow, the comment is update this these new values:\n\n```elixir\n%Comment{\n  id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n  message: \"This is (not) my first comment (anymore)\",\n  rating: 5,\n  timestamp: \"2024-07-23T12:00:00Z\"\n}\n```\n\nThe event structure will look like this:\n\n```elixir\n%WalEx.Event{\n  name: :comments,\n  type: :update,\n  source: %WalEx.Event.Source{\n    name: \"WalEx\",\n    version: \"4.4.0\",\n    db: \"postgresql\",\n    schema: \"public\",\n    table: \"comments\",\n    columns: %{\n      id: \"uuid\",\n      message: \"varchar\",\n      rating: \"integer\",\n      timestamp: \"datetime\"\n    }\n  },\n  new_record: %Comment{\n    id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n    message: \"This is (not) my first comment (anymore)\",\n    rating: 5,\n    timestamp: \"2024-07-23T12:00:00Z\"\n  },\n  old_record: nil,\n  changes: %{\n    message: %{\n      old_value: \"This is my first comment\",\n      new_value: \"This is (not) my first comment (anymore)\"\n    },\n    rating: %{\n      old_value: 4,\n      new_value: 5\n    }\n  },\n  timestamp: \"2024-08-25T13:13:30Z\",\n  lsn: {0, 0}\n}\n```","ref":"EventStreamex.Operators.Operator.html#t:event/0"},{"type":"module","title":"EventStreamex.Operators.ProcessStatus","doc":"Handle the process status of entities.\n\nEverytime an event is processed, it's saved in the `ProcessStatus` along\nwith the processed timestamp.\n\nThus, the process status contains the latest processed timestamp for all entities.\n\nYou can use this module to ensure an entity has been processed.\n\nLet's say you have `posts` and `post_with_comments_counts` entities.\nWhen a `posts` is created, an operator, listening for `posts` will create\na `post_with_comments_counts` entity with the same ID.\n\nYou have an API endpoint to create a new `posts` (`POST /posts`), and as soon as your appliation\ncalls it, it redirects to another view requesting the detail of the created `post_with_comments_counts` (`GET /post_with_comments_counts`).\n\nBut the creation of the later is asynchronous, so you may receive a 404 error from your API.\n\nIn order to avoid that, you could implement a mecanism that would return, along\nwith the `posts` newly created, the post creation timestamp.\n\nThen, in the `GET /post_with_comments_counts` endpoint, you would send along the timestamp you got\nfrom the `POST /posts` endpoint.\n\nAnd in your controller you would check if the entity has been\nprocessed before returning the `post_with_comments_counts` requested.\n\nYou can use `processed?/2` in order to do so.\n\nFor instance, when you create a `posts`, you receive this header from the API:\n\n```\nX-Event-Streamex-Entity-Timestamp: \"posts/1422057007123000\"\n```\n_(You can use the `:process_status_entity_field` config parameter to set\nthe field to use in your entities to know the last updated time.\nYou should return the field's value in your API, and use this value\nin the `processed?/2` function)_\n\nYou will send this header when you call the `GET /post_with_comments_counts` endpoint.\n\nAnd the controller will immediately check if the entity has been processed:\n\n```elixir\nEventStreamex.Operators.ProcessStatus.processed?(\"posts\", 1422057007123000)\n```\n\nIf the entity has not been processed yet, then, you know the `post_with_comments_counts` entity\nhas not been created yet (it works the same for updated entities).\nSo you can immediately return a 425 response for instance, informing the client to retry later.\n\nIf the entity has been processed, then you fetch it and return it.\nIf you get a 404 error, this time, that means the entity did not exist.","ref":"EventStreamex.Operators.ProcessStatus.html"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.ProcessStatus.html#child_spec/1"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.entity_processed/2","doc":"Updates the processed timestamp of an entity **(do not use this function yourself)**.","ref":"EventStreamex.Operators.ProcessStatus.html#entity_processed/2"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.last_processed/1","doc":"Returns the last timestamp when the specified entity has been processed","ref":"EventStreamex.Operators.ProcessStatus.html#last_processed/1"},{"type":"function","title":"Params - EventStreamex.Operators.ProcessStatus.last_processed/1","doc":"The entity to check","ref":"EventStreamex.Operators.ProcessStatus.html#last_processed/1-params"},{"type":"function","title":"Returns - EventStreamex.Operators.ProcessStatus.last_processed/1","doc":"The last time this entity was processed","ref":"EventStreamex.Operators.ProcessStatus.html#last_processed/1-returns"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.processed?/2","doc":"Has the entity with the specified timestamp already been processed?","ref":"EventStreamex.Operators.ProcessStatus.html#processed?/2"},{"type":"function","title":"Params - EventStreamex.Operators.ProcessStatus.processed?/2","doc":"* `entity`: The entity to check\n* `timestamp`: Usually the `update_at` field of the entity","ref":"EventStreamex.Operators.ProcessStatus.html#processed?/2-params"},{"type":"function","title":"Returns - EventStreamex.Operators.ProcessStatus.processed?/2","doc":"`true` if this entity with this timestamp, or a later timestamp has been processed.\n`false` otherwise.","ref":"EventStreamex.Operators.ProcessStatus.html#processed?/2-returns"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.reset/0","doc":"Removes all items from the process status.\n\n**Should only be used for testing purposes**","ref":"EventStreamex.Operators.ProcessStatus.html#reset/0"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.start_link/1","doc":"","ref":"EventStreamex.Operators.ProcessStatus.html#start_link/1"},{"type":"module","title":"EventStreamex.Operators.ProcessStatus.DbAdapter","doc":"A database queue adapter.\n\nThe queue is stored in the database.\nIt uses the information from the configuration to connect to the database\nand it automatically creates the database to store the queue.","ref":"EventStreamex.Operators.ProcessStatus.DbAdapter.html"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.DbAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.ProcessStatus.DbAdapter.html#child_spec/1"},{"type":"module","title":"EventStreamex.Operators.ProcessStatus.NoAdapter","doc":"Do not store the process status.\n\nThis adapter can be used to improve performances.\nThis adapter should only be used for tests or dev environments as it will\nnot recover the process status in case of a node crash in a distributed system.","ref":"EventStreamex.Operators.ProcessStatus.NoAdapter.html"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.NoAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.ProcessStatus.NoAdapter.html#child_spec/1"},{"type":"behaviour","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter","doc":"Behaviour for the process status storage.\n\nEverytime an event is received we will save its name and timestamp to\nkeep a process status up to date.\n\nThe process status allows us to know for each entity the last time it has been processed.\nHere, \"processed\" means all of these steps:\n- Event received from WAL\n- Event emitted in pubsub\n- Event processed by all operators listening to it\n\nThe adapters keeps a copy of this process status, mostly in case of crash (Except for `EventStreamex.Operators.ProcessStatus.NoAdapter` which does not saves the process status and should not be used in production)\n\nThere are currently 2 process status adapters:\n* `EventStreamex.Operators.ProcessStatus.DbAdapter`: A process status that uses the database tu store its items (This is the default process status adapter)\n* `EventStreamex.Operators.ProcessStatus.NoAdapter`: A process status that do not store its items\n\nYou can create your own adapter and set the config to use it.\n\nHere is the full code of a possible memory adapter:\n\n```elixir\ndefmodule EventStreamex.Operators.ProcessStatus.MemAdapter do\n  use GenServer\n  @behaviour EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter\n\n  @impl true\n  def start_link(opts) do\n    GenServer.start_link(__MODULE__, opts, name: __MODULE__)\n  end\n\n  @impl EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter\n  def item_processed(item) do\n    GenServer.call(__MODULE__, {:save, item})\n  end\n\n  @impl EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter\n  def load() do\n    GenServer.call(__MODULE__, :load)\n  end\n\n  @impl EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter\n  def reset() do\n    GenServer.call(__MODULE__, :reset)\n  end\n\n  # Callbacks\n\n  @impl true\n  def init(_opts) do\n    {:ok, %{}}\n  end\n\n  @impl true\n  def handle_call({:save, {entity, timestamp}}, _from, process_status) do\n    new_process_status = process_status |> Map.put(entity, timestamp)\n    {:reply, {:ok, new_process_status}, new_process_status}\n  end\n\n  @impl true\n  def handle_call(:load, _from, process_status) do\n    {:reply, {:ok, process_status |> Map.to_list()}, process_status}\n  end\n\n  @impl true\n  def handle_call(:reset, _from, _process_status) do\n    {:reply, {:ok, []}, []}\n  end\nend\n```\n\nThe adapter must be a `GenServer` and declare the function `c:start_link/1`.\n\nThen, it must provide these functions:\n* `c:item_processed/1`: Updates the timestamps of an entity (or adds it if the entity was not known before)\n* `c:load/0`: Return the entire process status\n* `c:reset/0`: Remove all items from the process status\n\nAn item is a tuple containing:\n* The name of the entity (as a string)\n* The timestamp of the last event processed for this entity","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html"},{"type":"function","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html#child_spec/1"},{"type":"callback","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.item_processed/1","doc":"Callback called when an entity has just been processed.\n\nThis callback must save it with its timestamps\n\nThe item is a `t:process_status/0`.\n\nThe return value is a result tuple.\nThe value in case of success is not used for the moment.","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html#c:item_processed/1"},{"type":"callback","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.load/0","doc":"Called at startup or in case of processor crash to return the full list of items.\n\nThe return value is a result tuple with the list of items `t:process_status/0`.","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html#c:load/0"},{"type":"callback","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.reset/0","doc":"Called for tests to reset the process status\n\nThe return value is a result tuple with an empty list.","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html#c:reset/0"},{"type":"callback","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.start_link/1","doc":"Called when starting the process status adapter.\n\nThe parameter is the one given in the config file when you set the adapter.\n\nThe return value is a result with the `pid` of the adapter processor.","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html#c:start_link/1"},{"type":"type","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.process_status/0","doc":"A process status item.\n\nAn item is a tuple containing:\n* The entity name\n* The timestamp of the last event processed for this entity","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html#t:process_status/0"},{"type":"type","title":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.process_status_list/0","doc":"A list of `t:process_status/0`","ref":"EventStreamex.Operators.ProcessStatus.ProcessStatusStorageAdapter.html#t:process_status_list/0"},{"type":"module","title":"EventStreamex.Operators.Queue","doc":"Handle the queue of operators to process.\n\nEvery event received is dispatched to operators (`EventStreamex.Operators.Operator`),\nand executed sequentially to keep ordenancing.\n\nEvery time an event is received, a task is added to the queue of events with each\noperator listening for this event.\nAnd everytime a an operator task is finished, it is update in the queue, until it is deleted completely when all operators have finished their task.\n\nA queue item is a tuple containing the operator modules with their completion status and the event to process.\n\nThe queue is handled internally.\nYou should not use it directly.","ref":"EventStreamex.Operators.Queue.html"},{"type":"function","title":"EventStreamex.Operators.Queue.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Queue.html#child_spec/1"},{"type":"function","title":"EventStreamex.Operators.Queue.enqueue/2","doc":"Adds another item to the queue","ref":"EventStreamex.Operators.Queue.html#enqueue/2"},{"type":"function","title":"EventStreamex.Operators.Queue.get_queue/0","doc":"Retrieves the full queue","ref":"EventStreamex.Operators.Queue.html#get_queue/0"},{"type":"function","title":"EventStreamex.Operators.Queue.get_task/0","doc":"Retrieves the current task","ref":"EventStreamex.Operators.Queue.html#get_task/0"},{"type":"function","title":"EventStreamex.Operators.Queue.reset_queue/0","doc":"Removes all items from the queue.\n\n**Should only be used for testing purposes**","ref":"EventStreamex.Operators.Queue.html#reset_queue/0"},{"type":"function","title":"EventStreamex.Operators.Queue.task_finished/1","doc":"Tells the queue that the module in the current item has finished.\n\nIt will update its completion state, and if all modules have finished\nthe task is completed","ref":"EventStreamex.Operators.Queue.html#task_finished/1"},{"type":"module","title":"EventStreamex.Operators.Queue.DbAdapter","doc":"A database queue adapter.\n\nThe queue is stored in the database.\nIt uses the information from the configuration to connect to the database\nand it automatically creates the database to store the queue.","ref":"EventStreamex.Operators.Queue.DbAdapter.html"},{"type":"function","title":"EventStreamex.Operators.Queue.DbAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Queue.DbAdapter.html#child_spec/1"},{"type":"module","title":"EventStreamex.Operators.Queue.NoAdapter","doc":"Adapter that does nothing with the queue.\n\nThis adapter simply improves performances.\nThis adapter should only be used for tests or dev environments as it will\nnot recover the queue in case of a node crash in a distributed system.","ref":"EventStreamex.Operators.Queue.NoAdapter.html"},{"type":"function","title":"EventStreamex.Operators.Queue.NoAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Queue.NoAdapter.html#child_spec/1"},{"type":"behaviour","title":"EventStreamex.Operators.Queue.QueueStorageAdapter","doc":"Behaviour for the queue storage.\n\nEvery event received is dispatched to operators (`EventStreamex.Operators.Operator`),\nand executed sequentially to keep ordenancing.\n\nEvery time an event is received, a task is added to the queue of events with each\noperator listening for this event.\nAnd everytime a an operator task is finished, its completion status is updated in the task.\nWhen all operators have finished their task for this event, it is removed from the queue.\n\nThe adapters keeps a copy of this queue, mostly in case of crash (Except for `EventStreamex.Operators.Queue.NoAdapter` which does not save the queue and should not be used in production)\n\nThere are currently 2 queue adapters:\n* `EventStreamex.Operators.Queue.DbAdapter`: A queue that uses the database tu store its items (This is the default queue adapter)\n* `EventStreamex.Operators.Queue.NoAdapter`: A queue that does not store its items\n\nYou can create your own adapter and set the config to use it.\n\nHere is the full code of a possible memory adapter:\n\n```elixir\ndefmodule EventStreamex.Operators.Queue.MemAdapter do\n  use GenServer\n  @behaviour EventStreamex.Operators.Queue.QueueStorageAdapter\n\n  @impl true\n  def start_link(opts) do\n    GenServer.start_link(__MODULE__, opts, name: __MODULE__)\n  end\n\n  @impl EventStreamex.Operators.Queue.QueueStorageAdapter\n  def add_item(item) do\n    GenServer.call(__MODULE__, {:save, item})\n  end\n\n  @impl EventStreamex.Operators.Queue.QueueStorageAdapter\n  def delete_item(item) do\n    GenServer.call(__MODULE__, {:delete, item})\n  end\n\n  @impl EventStreamex.Operators.Queue.QueueStorageAdapter\n  def update_processors_status(item) do\n    GenServer.call(__MODULE__, {:update, item})\n  end\n\n  @impl EventStreamex.Operators.Queue.QueueStorageAdapter\n  def load_queue() do\n    GenServer.call(__MODULE__, :load)\n  end\n\n  @impl EventStreamex.Operators.Queue.QueueStorageAdapter\n  def reset_queue() do\n    GenServer.call(__MODULE__, :reset)\n  end\n\n  # Callbacks\n\n  @impl true\n  def init(_opts) do\n    {:ok, []}\n  end\n\n  @impl true\n  def handle_call({:save, item}, _from, queue) do\n    new_queue = queue ++ [item]\n    {:reply, {:ok, new_queue}, new_queue}\n  end\n\n  @impl true\n  def handle_call({:delete, {id, _item}}, _from, queue) do\n    new_queue = queue |> Enum.reject(fn {i, _t} -> i == id end)\n    {:reply, {:ok, new_queue}, new_queue}\n  end\n\n  @impl true\n  def handle_call({:update, {id, item}}, _from, queue) do\n    new_queue = queue |> Enum.map(fn {i, t} ->\n      if i == id do\n        item\n      else\n        t\n      end\n    end)\n    {:reply, {:ok, new_queue}, new_queue}\n  end\n\n  @impl true\n  def handle_call({_action, nil}, _from, queue) do\n    {:reply, {:ok, queue}, queue}\n  end\n\n  @impl true\n  def handle_call(:load, _from, queue) do\n    {:reply, {:ok, queue}, queue}\n  end\n\n  @impl true\n  def handle_call(:reset, _from, _queue) do\n    {:reply, {:ok, []}, []}\n  end\nend\n```\n\nThe adapter must be a `GenServer` and declare the function `c:start_link/1`.\n\nThen, it must provide these functions:\n* `c:add_item/1`: Add a new item to the queue\n* `c:delete_item/1`: Delete an item from the queue\n* `c:update_processors_status/1`: Updates the specified item\n* `c:load_queue/0`: Return the entire queue\n* `c:reset_queue/0`: Remove all items from the queue\n\nAn item is a tuple containing:\n* The ID of the queue item\n* The operator modules that process the event. It's a map with the module of the operator as key and a boolean as value. The boolean indicates the completion status for this operator\n* The WalEx event (more on this structure in `EventStreamex.EventListener`)","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html"},{"type":"callback","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.add_item/1","doc":"Callback called when a new item must be stored in the queue.\n\nThe item is a `t:queue_item/0`.\n\nThe return value is a result tuple.\nThe value in case of success is not used for the moment.","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#c:add_item/1"},{"type":"function","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#child_spec/1"},{"type":"callback","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.delete_item/1","doc":"Callback called when an item has been processed and must be deleted from the queue.\n\nThe item is a `t:queue_item/0`.\n\nThe return value is a result tuple.\nThe value in case of success is not used for the moment.","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#c:delete_item/1"},{"type":"callback","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.load_queue/0","doc":"Called at startup or in case of processor crash to return the full list of items.\n\nThe return value is a result tuple with the list of items `t:queue/0`.","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#c:load_queue/0"},{"type":"callback","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.reset_queue/0","doc":"Called for tests to reset the queue\n\nThe return value is a result tuple with an empty list.","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#c:reset_queue/0"},{"type":"callback","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.start_link/1","doc":"Called when starting the queue adapter.\n\nThe parameter is the one given in the config file when you set the adapter.\n\nThe return value is a result with the `pid` of the adapter processor.","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#c:start_link/1"},{"type":"callback","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.update_processors_status/1","doc":"Updates the first item in the queue.\n\nThis is used to update the processors status from the current task","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#c:update_processors_status/1"},{"type":"type","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.queue/0","doc":"A list of `t:queue_item/0`","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#t:queue/0"},{"type":"type","title":"EventStreamex.Operators.Queue.QueueStorageAdapter.queue_item/0","doc":"A queue item to store.\n\nAn item is a tuple containing:\n* The ID of the queue item\n* The operator modules that process the event. It's a map with the module of the operator as key and a boolean as value. The boolean indicates the completion status for this operator\n* The WalEx event (more on this structure in `EventStreamex.EventListener`)","ref":"EventStreamex.Operators.Queue.QueueStorageAdapter.html#t:queue_item/0"},{"type":"module","title":"EventStreamex.Operators.Scheduler","doc":"Is responsible for starting operators as soon as an event is received.\n\nWhen an event is received, the scheduler processes it by searching for\nany `EventStreamex.Operators.Operator` listening to it.\n\nThe it will start an `EventStreamex.Operators.Executor` who will be responsible\nfor the completion of the operator.\n\nIf the operator fails too many times, then, the executor will also fail,\nas well as the scheduler.\n\nIf this happens that means that no more event will be processed.\nThis is a security to avoid inconsistency in your data, as events must be\nprocessed in order.\n\nIf the scheduler fails, you will have to restart it yourself using `EventStreamex.restart_scheduler()`.\n\nBut before you do that, ensure that the operator will not crash again. Maybe it will\nrequire some code changes. Or a system to be up and running again.\n\nDo not hesitate to implement your own `EventStreamex.Operators.Logger.ErrorLoggerAdapter`\nto be notified why a crash appeared.","ref":"EventStreamex.Operators.Scheduler.html"},{"type":"function","title":"EventStreamex.Operators.Scheduler.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Operators.Scheduler.html#child_spec/1"},{"type":"function","title":"EventStreamex.Operators.Scheduler.curr_job/0","doc":"Gets the list of operators being executed with their status.\n\nThe return value is like this: `[{atom(), {pid(), ref(), boolean()}}]`","ref":"EventStreamex.Operators.Scheduler.html#curr_job/0"},{"type":"function","title":"EventStreamex.Operators.Scheduler.curr_job/1","doc":"Gets the information of the given operator being executed or `:no_job`.\n\nThe return value is like this: `{pid(), ref(), event(), boolean()}`","ref":"EventStreamex.Operators.Scheduler.html#curr_job/1"},{"type":"function","title":"EventStreamex.Operators.Scheduler.is_alive?/0","doc":"Returns the status of the scheduler.","ref":"EventStreamex.Operators.Scheduler.html#is_alive?/0"},{"type":"function","title":"EventStreamex.Operators.Scheduler.process_event/2","doc":"Function used to process an event.\n\n**You must not use this function yourself**","ref":"EventStreamex.Operators.Scheduler.html#process_event/2"},{"type":"function","title":"Event example: - EventStreamex.Operators.Scheduler.process_event/2","doc":"```\n%Walex.Event{\n  name: :user,\n  type: :update,\n  source: %WalEx.Event.Source{\n    name: \"WalEx\",\n    version: \"3.8.0\",\n    db: \"todos\",\n    schema: \"public\",\n    table: \"user\",\n    columns: %{\n      id: \"integer\",\n      name: \"varchar\",\n      created_at: \"timestamptz\"\n    }\n  },\n  new_record: %{\n    id: 1234,\n    name: \"Chase Pursley\",\n    created_at: #DateTime<2023-08-18 14:09:05.988369-04:00 -04 Etc/UTC-4>\n  },\n  # we don't show old_record for update to reduce payload size\n  # however, you can see any old values that changed under \"changes\"\n  old_record: nil,\n  changes: %{\n    name: %{\n      new_value: \"Chase Pursley\",\n      old_value: \"Chase\"\n    }\n  },\n  timestamp: ~U[2023-12-18 15:50:08.329504Z]\n}\n```","ref":"EventStreamex.Operators.Scheduler.html#process_event/2-event-example"},{"type":"module","title":"EventStreamex.Orchestrator","doc":"Module responsible for launching the event streaming system.\n\nThe event streaming system cannot run at the same time in different nodes.\nOtherwise it would crash because of a duplicate replica slot.\nAnd if we changed the name of the replica slot for each node it would, then,\nduplicate all WAL events received by PostgreSQL.\n\nThis is something we don't want, thus, we have this orchestrator that will\nensure only one instance of the system is running at the same time.\n\nIt will also ensure that if the master node (the one running the system) crashes or terminates,\nthe system will be started on another node to continue processing new WAL events.\n\nEvery node runs the orchestrator.\nWhen a node starts it generates a unique identifier based on the current time\nand communicates that id to the other nodes.\n\nWhen a node terminates a new election is run based on the lowest node identifier.\nThus, all nodes will agree on the new master node.","ref":"EventStreamex.Orchestrator.html"},{"type":"function","title":"EventStreamex.Orchestrator.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EventStreamex.Orchestrator.html#child_spec/1"},{"type":"function","title":"EventStreamex.Orchestrator.supervisor_alive?/0","doc":"Checks if the supervisor is alive.\n\nThe supervisor is the event streaming system.\nIt should be alive if the current node is the master node.","ref":"EventStreamex.Orchestrator.html#supervisor_alive?/0"},{"type":"extras","title":"EventStreamex","doc":"# EventStreamex\n\nPackage to add event streaming in your Elixir application, using PostgreSQL WAL.\n\nWorks very well with [Phoenix LiveView](https://hexdocs.pm/phoenix_live_view/welcome.html) to add realtime to your app.\n\nThis package uses the [Write-Ahead-Log from PostgreSQL](https://www.postgresql.org/docs/current/wal-intro.html) to listen for changes in database tables and then, handle them in your application in 2 ways:\n\n- To add realtime support in your views\n- To update other entities in realtime (as soon as an entity it depends on is updated)\n\nFull documentation can be found [here](https://hexdocs.pm/event_streamex).","ref":"readme.html"},{"type":"extras","title":"Usage - EventStreamex","doc":"","ref":"readme.html#usage"},{"type":"extras","title":"Installation - EventStreamex","doc":"The package can be installed by adding `event_streamex` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:event_streamex, \"~> 1.3.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Configuration - EventStreamex","doc":"Here are the configuration parameters for `EventStreamex` with their default value:\n\n```elixir\nconfig :event_streamex,\n  app_name: nil,\n  database: \"postgres\",\n  durable_slot: true,\n  error_logger_adapter: {EventStreamex.Operators.Logger.LoggerAdapter, []},\n  hostname: \"localhost\",\n  operator_queue_backoff_multiplicator: 2,\n  operator_queue_max_restart_time: 10000,\n  operator_queue_max_retries: 5,\n  operator_queue_min_restart_time: 500,\n  password: \"postgres\",\n  port: \"5432\",\n  process_status_entity_field: :updated_at,\n  process_status_storage_adapter: {EventStreamex.Operators.ProcessStatus.DbAdapter, []},\n  publication: \"events\",\n  pubsub: [adapter: Phoenix.PubSub, name: nil],\n  queue_storage_adapter: {EventStreamex.Operators.Queue.DbAdapter, []},\n  slot_name: \"postgres_slot\",\n  url: \"\",\n  username: \"postgres\"\n```\n\n- `app_name`: Your application name\n- `database`: Name of the database to connect to\n- `durable_slot`: Is the replication slot for the WAL temporary or durable? If you set it to `false`, then, you will loose all WAL events sent while your application was not running.\n- `error_logger_adapter`: The logger module used with the `start_link` parameter to send it\n- `hostname`: Your PostgreSQL hostname\n- `operator_queue_backoff_multiplicator`: Multiplicator used to compute the new wait time after an operator crash (See `EventStreamex.Operators.Logger.ErrorLoggerAdapter`)\n- `operator_queue_max_restart_time`: Maximum time to wait in milliseconds before the restart of a crashed operator (See `EventStreamex.Operators.Logger.ErrorLoggerAdapter`)\n- `operator_queue_max_retries`: Maximum number of restarts of a crashed operator (See `EventStreamex.Operators.Logger.ErrorLoggerAdapter`)\n- `operator_queue_min_restart_time`: Time to wait in milliseconds before the first restart of a crashed operator (See `EventStreamex.Operators.Logger.ErrorLoggerAdapter`)\n- `password`: Password used to connect to the database\n- `port`: Port of the database\n- `process_status_entity_field`: The field name used is your entities to represent the last updated time. It must be set also when the entity is created. And it should have a precision of microseconds (use the `utc_datetime_usec` type with `Ecto`). This field is used to update the `EventStreamex.Operators.ProcessStatus` registry.\n- `process_status_storage_adapter`: Module used for process status storage and the parameter sent to the `start_link` function\n- `publication`: Name of the publication used by the WAL (it is created automatically if it does not already exist)\n- `pubsub`: Adapter to use for pubsub and its name\n- `queue_storage_adapter`: Module used for queue storage and the parameter sent to the `start_link` function\n- `slot_name`: Name of the replication slot for the WAL. It is automatically created if it does not exist.\n- `url`: URL to connect to the database. If it is not set, the other parameters will be used instead\n- `username`: Username to connect to the database\n\nFor the package to work correctly you will have to specify, at least:\n\n- The database connection parameters\n- The `:app_name`\n- The `pubsub` `:name`","ref":"readme.html#configuration"},{"type":"extras","title":"Mark the entity for listening - EventStreamex","doc":"To mark an entity for listening, you will have to do 2 things:\n\n- use the `EventStreamex.Events` module\n- and set the tables for replication\n\n#### Use the `EventStreamex.Events` module\n\n```elixir\ndefmodule MyApp.Blog.Post do\n  use MyApp.Schema\n\n  use EventStreamex.Events, name: :my_app, schema: \"posts\"\n\n  import Ecto.Changeset\n\n  alias MyApp.Blog.PostWithCommentsCount\n\n  schema \"posts\" do\n    field :title, :string\n    field :body, :string\n\n    timestamps(type: :utc_datetime)\n  end\n\n  @doc false\n  def changeset(post, attrs) do\n    post\n    |> cast(attrs, [:id, :title, :body])\n    |> validate_required([:title, :body])\n  end\nend\n```\n\n#### Set the tables for replication\n\nFor each table of your application, you will have to execute this SQL query (ideally in the migrations);\n\n```sql\nALTER TABLE ONLY   REPLICA IDENTITY FULL;\n```\n\nOf course, you will replace ` `by the name of the table.\n\nAny marked entity will tell PostreSQL's WAL to emit events as soon as it is `inserted`, `updated`, `deleted`.","ref":"readme.html#mark-the-entity-for-listening"},{"type":"extras","title":"Listen to events in live views - EventStreamex","doc":"Now that our `posts` entity is subscribed to, in the WAL, we can listen to its changes in a live view:\n\n```elixir\ndefmodule MyApp.PostLive.Index do\n  use MyApp, :live_view\n\n  use EventStreamex.EventListener,\n    schema: \"post_with_comments_counts\",\n    subscriptions: [:unscoped]\n\n  alias MyApp.Blog\n  alias MyApp.Blog.Post\n\n  @impl true\n  def mount(params, session, socket) do\n    {:ok, socket} = super(params, session, socket)\n\n    {:ok, stream(socket, :posts, Blog.list_posts())}\n  end\n\n  @impl true\n  def handle_params(params, url, socket) do\n    {_res, socket} = super(params, url, socket)\n\n    {:noreply, socket}\n  end\n\n  @impl true\n  def handle_info({:on_insert, [], \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> stream_insert(:posts, post.new_record)\n  end\n\n  @impl true\n  def handle_info({:on_update, [], \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> stream_insert(:posts, post.new_record)\n  end\n\n  @impl true\n  def handle_info({:on_delete, [], \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> stream_delete(:posts, post.old_record)\n  end\nend\n```\n\nHere, we listen to `posts` events in our live view.\n\nThe `handle_info/2` callbacks are called with the kind of event received (`:on_insert`, `:on_update`, `:on_delete`) and the event.\n\nYou can, then do whatever you need to update your view.","ref":"readme.html#listen-to-events-in-live-views"},{"type":"extras","title":"Events structure - EventStreamex","doc":"An event looks like this:\n\n```elixir\n%WalEx.Event{\n  name: atom(),\n  type: :insert | :update | :delete,\n  source: %WalEx.Event.Source{\n    name: String.t(),\n    version: String.t(),\n    db: String.t(),\n    schema: String.t(),\n    table: String.t(),\n    columns: map()\n  },\n  new_record: map() | nil,\n  old_record: map() | nil,\n  changes: map() | nil,\n  timestamp: DateTime.t(),\n  lsn: {integer(), integer()}\n}\n```\n\n_I am using the [WalEx package internally](https://github.com/cpursley/walex)_\n\n- `name`: The name of the entity (ie: `:comments` for a table named `comments`)\n- `type`: The type of event between `insert`, `update` and `delete`\n- `source`: Information about the event:\n  - `name`: \"WalEx\"\n  - `version`: Current version of `WalEx`\n  - `db`: The name of the database\n  - `schema`: Mostly `\"public\"`\n  - `table`: The name of the table (ie: `\"comments\"`)\n  - `columns`: A map of fields with their type (ie: `%{\"id\": \"integer\", \"message\": \"varchar\"}`)\n- `new_record`: The entity itself for `insert` and `update` events. `nil` for `delete` events.\n- `old_record`: The entity itself for `delete` events. `nil` for `insert` and `update` events.\n- `changes`: A map with the changes in the entity in `update` events, `nil` otherwise (see below)\n- `timestamp`: The timstamp of the event in `DateTime` type\n- `lsn`: A tuple containing information about the publication cursor\n\n#### `changes`\n\nWhen you receive an `update` event, you will also have the `changes` field set to a map containing\nthe changes the entity received since the update.\n\nThis map contains the changed fields as keys, and a map describing the change as value.\nThis \"change\" map contains 2 fields:\n\n- `old_value`: The value before the update\n- `new_value`: The value after the update\n\nFor instance, let's say you have a `comments` entity with 4 fields: `id`, `message`, `rating`, `timestamp`.\n\nYou have a comment with these values:\n\n```elixir\n%Comment{\n  id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n  message: \"This is my first comment\",\n  rating: 4,\n  timestamp: \"2024-07-23T12:00:00Z\"\n}\n```\n\nNow, the comment is update this these new values:\n\n```elixir\n%Comment{\n  id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n  message: \"This is (not) my first comment (anymore)\",\n  rating: 5,\n  timestamp: \"2024-07-23T12:00:00Z\"\n}\n```\n\nThe event structure will look like this:\n\n```elixir\n%WalEx.Event{\n  name: :comments,\n  type: :update,\n  source: %WalEx.Event.Source{\n    name: \"WalEx\",\n    version: \"4.4.0\",\n    db: \"postgresql\",\n    schema: \"public\",\n    table: \"comments\",\n    columns: %{\n      id: \"uuid\",\n      message: \"varchar\",\n      rating: \"integer\",\n      timestamp: \"datetime\"\n    }\n  },\n  new_record: %Comment{\n    id: \"dd4bc2ba-c7cc-4a05-a1c7-9f26cd9ab79f\",\n    message: \"This is (not) my first comment (anymore)\",\n    rating: 5,\n    timestamp: \"2024-07-23T12:00:00Z\"\n  },\n  old_record: nil,\n  changes: %{\n    message: %{\n      old_value: \"This is my first comment\",\n      new_value: \"This is (not) my first comment (anymore)\"\n    },\n    rating: %{\n      old_value: 4,\n      new_value: 5\n    }\n  },\n  timestamp: \"2024-08-25T13:13:30Z\",\n  lsn: {0, 0}\n}\n```","ref":"readme.html#events-structure"},{"type":"extras","title":"Update other entities based on events - EventStreamex","doc":"Now, let's assume that we also have a `comments` entity linked to a post,\nand that we want to display our posts with the number of comments they have.\n\nIn event streaming we will create another entity (let's say `post_with_comments_count`), and update it when we receive events from `posts` and `comments`.\n\nThis entity will look like this:\n\n- `id`: It's id (same one as the post id)\n- `title`: Same one as the post title\n- `body`: Same one as the post body\n- `comments_count`: The number of comments of the post\n\n```elixir\ndefmodule MyApp.Operators.PostWithCommentsCountOperator do\n  use EventStreamex.Operators.Operator,\n    schema: \"post_with_comments_counts\",\n    repo: MyApp.Repo\n\n  on_event(\n    all_events(),\n    entities([\"comments\", \"posts\"]),\n    query(\n      \"\"\"\n      MERGE INTO post_with_comments_counts dest\n      USING (\n          SELECT\n              p.id AS id,\n              p.title AS title,\n              p.body AS body,\n              p.inserted_at AS inserted_at,\n              p.updated_at AS updated_at,\n              (SELECT COUNT(c.id) FROM comments c WHERE c.post_id = p.id) AS comments_count\n          FROM posts p\n          WHERE p.id = $1\n      ) src ON src.id = dest.id\n      WHEN NOT MATCHED THEN\n        INSERT (id, title, body, inserted_at, updated_at, comments_count)\n        VALUES (src.id, src.title, src.body, src.inserted_at, src.updated_at, src.comments_count)\n      WHEN MATCHED THEN\n        UPDATE SET\n            title = src.title,\n            body = src.body,\n            inserted_at = src.inserted_at,\n            updated_at = src.updated_at,\n            comments_count = src.comments_count;\n      \"\"\",\n      id([{\"comments\", :post_id}, {\"posts\", :id}]),\n      delete_with([\"posts\"])\n    )\n  )\nend\n```\n\nUsing an operator will allow us to insert/update/delete an entity based on events.\n\nHere, we listen to `posts` and `comments` and, for each event, we execute the `MERGE` query.\n\nThis query will create the `post_with_comments_count` when a post is created, and update it when the post is updated.\nIt will also update the `comments_count` each time there is a new comment,\nor a comment has been deleted.\n\nThe `$1` parameter in the query will be replaced by the field in the listened entities that corresponds to the id of the `post_with_comments_count`.\nIn this example, it's the post ID. Which is found in the `post_id` field in `comments`.\nThat comes from this line: `id([{\"comments\", :post_id}, {\"posts\", :id}])`.\n\n`delete_with([\"posts\"])` will tell the operator that a post deletion must also delete the `post_with_comments_count`.\n\nIf you can more complex scenarios, you can also use a function instead of `query/3`:\n\n```elixir\non_event(\n  all_events(),\n  entities([\"comments\", \"posts\"]),\n  fn event ->\n    {:ok, res} =\n      event\n      |> MyQuery.do_some_queries()\n      |> MyQuery.then_another_query()\n    {:ok, res}\n  end\n)\n```","ref":"readme.html#update-other-entities-based-on-events"},{"type":"extras","title":"Why using event streaming? - EventStreamex","doc":"Although using event streaming with `EventStreamex` looks easy, why bother using event streaming in the first place?","ref":"readme.html#why-using-event-streaming"},{"type":"extras","title":"The issue without event streaming - EventStreamex","doc":"Let's say you have a `posts` entity which contains:\n\n- `title`: a title\n- `body`: a body\n\nYou also have a `comments` entity which contains:\n\n- `post_id`: The ID of the post it is attached to\n- `comment`: The comment content\n\nWhen you show a list of your posts you might want to show the number\nof comments it contains.\n\nTo do so you will be tempted to make a SQL query with a subquery that will\ncount the comments of your posts.\n\nThis is pretty trivial:\n\n```sql\nSELECT\n  p.id AS id,\n  p.title AS title,\n  p.body AS body,\n  (SELECT COUNT(*) FROM comments c WHERE c.post_id = p.id) AS comments_count\nFROM posts p;\n```\n\nBut it has several flaws.\nLet's see them in more details.\n\n#### Your queries will become slower and slower over time\n\nBecause subqueries are not free. They take more time to execute that a simple query with no join and subquery.\n\nSo, the more posts and comments you have, the slower the query will be.\n\n#### You will end up adding complexity in your queries\n\nAs your app evolves, you will probably create new entities related to your posts and comments.\nThus, you will add more and more subqueries and joins to your queries, making them slower and harder to maintain...\n\n#### Realtime will become a mess to maintain\n\nIf you want to handle realtime events like updating the number of comments when a new one is created it will be a mess to maintain overtime because you will end up subscribing to a lot of events as you link your posts to other entities.\n\nAnd you will probably forget to subscribe to some of the events making your app inconsistant (some features will be in realtime and other won't be).\n\n#### Almost impossible to maintain recursivity\n\nIf your comments can have subcomments it will be almost impossible\nto maintain the comments count because of the recursivity it implies.\n\n#### Depending on entities not directly related to the base entity\n\nLet's say you create another entity named `ratings` which saves each rating given to a specific comment.\n\nIt would look like this:\n\n- `comment_id`: The ID of the comment related to this rating\n- `rating`: The given rating\n\nNow imagine you want to display, for every comment, the average rating and the total count of ratings for this comment.\n\nHere again, you could solve this with some subqueries.\n\nBut what if you want to show the average rating of all comments in the post\nso that users can see that a post has received a lot of useful comments.\n\nI don't even want to think about the query that would require...","ref":"readme.html#the-issue-without-event-streaming"},{"type":"extras","title":"Benefits from using event streaming - EventStreamex","doc":"Let's come back to our `posts` and `comments`.\nNow let's imagine that instead of our subquery to count the comments we use\nan event streaming architecture and create another entity whose goal is\nspecifically to show the posts with their comments count.\n\nThis `post_with_comments_counts` entity will look like that:\n\n- `title`: a title\n- `body`: a body\n- `comments_count`: the number of comments of the post\n\nAnd this entity will listen to new comments to update its comments count:\nwhen a new comment is created it will increase the number of comments of the related post by 1.\n\nWith `EventStreamex` it would look like this:\n\n```elixir\ndefmodule MyApp.Operators.PostWithCommentsCountOperator do\n  use EventStreamex.Operators.Operator,\n    schema: \"post_with_comments_counts\",\n    repo: MyApp.Repo\n\n  on_event(\n    all_events(),\n    entities([\"comments\", \"posts\"]),\n    query(\n      \"MERGE INTO post_with_comments_counts dest\n      USING (\n          SELECT\n              p.id AS id,\n              p.title AS title,\n              p.body AS body,\n              p.inserted_at AS inserted_at,\n              p.updated_at AS updated_at,\n              (SELECT COUNT(c.id) FROM comments c WHERE c.post_id = p.id) AS comments_count\n          FROM posts p\n          WHERE p.id = $1\n      ) src ON src.id = dest.id\n      WHEN NOT MATCHED THEN\n        INSERT (id, title, body, inserted_at, updated_at, comments_count)\n        VALUES (src.id, src.title, src.body, src.inserted_at, src.updated_at, src.comments_count)\n      WHEN MATCHED THEN\n        UPDATE SET\n            title = src.title,\n            body = src.body,\n            inserted_at = src.inserted_at,\n            updated_at = src.updated_at,\n            comments_count = src.comments_count;\",\n      id([{\"comments\", :post_id}, {\"posts\", :id}]),\n      delete_with([\"posts\"])\n    )\n  )\nend\n```\n\nHere, we use the `EventStreamex.Operators.Operator` module to listen for\n`posts` and `comments` in all kinds of events (insert, update, delete).\n\nAnd each time a post or a comment is inserted, updated, deleted, we perform this `MERGE` query to create/update our `post_with_comments_counts`.\n\nThe `MERGE` query will insert a new post if the post does not exist and update it if it already exists.\n\n`id/1` is used to map the fields of the listened entities to the id of the\n`post_with_comments_counts` entity (`id` for a post and `post_id` for a comment).\n\nFinally, `delete_with/1` tells the operator which entities deletion should\nalso delete the related `post_with_comments_counts` entity.\n\nAs you can see, this `MERGE` query also uses a subquery, so why is it better than our first implementation?\n\n#### This is a naive implementation\n\nThe `query/3` macro is provided for simplicity.\nBut you can also replace it by a function to do things as you think it's best (optimize the queries according to the type of event and entity received).\n\nThis function will receive the event, allowing you to do whatever you want with it.\n\n#### The query is executed in the background\n\nIn our first example, the query was executed everytime someone wants to display the list of posts.\n\nIn our event streaming example, the query is executed in the background each time there is an event with a post or a comment.\n\nTo display your list of posts you will just have to do:\n\n```sql\nSELECT * FROM post_with_comments_counts;\n```\n\nWhich is simpler and faster than our former query.\n\n#### Realtime is easy\n\nYou can use the module `EventStreamex.EventListener` to easily add realtime\nin your live view applications.\n\nBecause we are using PostgreSQL WAL events, everytime something happens on an entity (`post_with_comments_counts` as well), an event is sent.\n\nThis event will be processed by operators listening to it (like our operator here does for `posts` and `comments`), and it will also be sent to a pubsub processor.\n\nThe `EventStreamex.EventListener`, now just has to listen to this pubsub processor to notify your live view that the entity your are showing as been updated.\n\nEverything, now, can be realtime, without beeing hard to maintain.\n\n#### A better pattern to model your data\n\nWith this approach you keep your entities consistent and simple.\nAnd you create specialized entities for what you want to show in your views.\n\nA post will stay a post with its `title` and `body` fields and a comment will stay a comment with a `comment` field.\n\n`post_with_comments_count` is just a specialized entity used for your view.\n\nThis keeps your data and APIs consistent and simple with explicit fields and names.\n\nFor instance, let's say you have a `users` entity with a `login` and a `password` field.\nBut later on, you realize that you need to store some information about your users, like its name, its address, and so on.\n\nYou will be tempted to update your `users` entity to add these new fields:\n\n- `login`\n- `password`\n- `address`\n- `name`\n\nBut still, in your application, when comes the moment to create your user, you only need the `login` and `password`.\n\nSo what do you do? You set the other fields as optionals in your API and set them a default value (empty string or null).\n\nAnd in the view that allows a user to specify more information about him (`name` and `address`), you will either:\n\n- Use the same API endpoint as for the `users` creation\n- Or create a new dedicated API endpoint\n\nBut both approachs are bad practices!\n\nIn the first approach that means that your API expects 2 mandatory fields: `login` and `password`, and 2 optional fields: `name` and `address`.\nBut here, there is no sense in sending the `login` and `password`.\nAnd maybe you would like the `name` and `address` fields to be mandatory.\nBut because your API set these fields as optionals, you will have to check for them from the client side.\nWhich is a bad practice.\n\nIn the second approach, that means that you will create an API endpoint for the user profile.\nBut this API, will finally update the `users` entity.\nYou end up with hard to understand APIs because their name does not match\nto the entity being modified.\n\nA better approach would be to have a `users` entity with:\n\n- `login`\n- `password`\n\nAnd its own API endpoint `/users` with only these two fields.\n\nAnd you would have a `user_profiles` entity with:\n\n- `user_id`\n- `name`\n- `address`\n\nAnd its own API endpoint `/users/{user_id}/user_profiles` with only the `name` and `address` fields.\n\nAnd what if you need to display all of these information in your view?\nWell, you use an operator like we did for the `post_with_comments_counts`.\nNamed, for instance: `user_with_profiles`, and listening to `users` and `user_profiles` events.\n\nThis way, you keep your data and API consistent.","ref":"readme.html#benefits-from-using-event-streaming"},{"type":"extras","title":"How to handle asynchronicity of entity creation - EventStreamex","doc":"Event streaming is nice, but it makes create/update/delete operations asynchronous for derived entities.\nWhich can be a nightmare to handle correctly.\n\nLet's say you have a LiveView application showing a list of `post_with_comments_counts`, which is a derived entity from the `posts` entity, containing, in addition to `posts` fields, a `comments_count` field.\n\nFrom this view, we can also create a new post.\nBut we don't create a `post_with_comments_counts` directly.\nInstead, we must create a `posts` entity.\nWhich, in turn, will trigger the creation of the `post_with_comments_counts` using event streaming.\n\nWhat we want to do is to redirect to the post detail page when the post is\ncreated.\n\nBut we have a problem!\n\nBecause the creation of the `posts` entity is synchronous, we can redirect\nas soon as it is created.\nBut that does not mean the `post_with_comments_counts` entity will be created already... while it's this entity we need to show.\n\nTo handle this scenario and be able to display the `post_with_comments_counts` entity as soon as it is created we have 3 solutions:","ref":"readme.html#how-to-handle-asynchronicity-of-entity-creation"},{"type":"extras","title":"Wait for the entity to be created before redirecting - EventStreamex","doc":"To ensure you redirect to an entity that exists, you can actualy wait for it\nto be created.\n\nIn order to do so, you will have to listen to this entity to know when it's been created.\n\nYou can use the `subscribe_entity/4` function for that matter.\n\n```elixir\ndefmodule MyApp.PostLive.Index do\n  use MyApp, :live_view\n\n  use EventStreamex.EventListener,\n    schema: \"post_with_comments_counts\",\n    subscriptions: [:unscoped]\n\n  alias MyApp.Blog\n  alias MyApp.Blog.Post\n\n  @impl true\n  def mount(params, session, socket) do\n    {:ok, socket} = super(params, session, socket)\n\n    {:ok, stream(socket, :posts, Blog.list_posts())}\n  end\n\n  @impl true\n  def handle_params(params, url, socket) do\n    {_res, socket} = super(params, url, socket)\n\n    {:noreply, apply_action(socket, socket.assigns.live_action, params)}\n  end\n\n  defp apply_action(socket, :new, _params) do\n    socket\n    |> assign(:page_title, \"New Post\")\n    |> assign(:post, %Post{})\n  end\n\n  defp apply_action(socket, :index, _params) do\n    socket\n    |> assign(:page_title, \"Listing Posts\")\n    |> assign(:post, nil)\n  end\n\n  @impl true\n  def handle_info({MyApp.PostLive.FormComponent, {:saved, post}}, socket) do\n    {:noreply,\n     socket\n     |> subscribe_entity(\"post_with_comments_counts\", :direct, %{\"id\" => post.id})}\n  end\n\n  @impl true\n  def handle_info({:on_insert, :direct, \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> unsubscribe_entity(\"post_with_comments_counts\", :direct)\n     |> push_navigate(to: \"/posts/#{post.new_record.id}\")}\n  end\n\n  @impl true\n  def handle_info({_, :direct, \"post_with_comments_counts\", _post}, socket) do\n    {:noreply, socket}\n  end\n\n  @impl true\n  def handle_info({:on_insert, [], \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> stream_insert(:posts, post.new_record)\n  end\n\n  @impl true\n  def handle_info({:on_update, [], \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> stream_insert(:posts, post.new_record)\n  end\n\n  @impl true\n  def handle_info({:on_delete, [], \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> stream_delete(:posts, post.old_record)\n  end\nend\n```\n\nHere, our live view listens for `:unscoped` `post_with_comments_counts` entity for the `index` view in realtime (nothing new).\n\nThis code is more interesting:\n\n```elixir\n@impl true\ndef handle_info({MyApp.PostLive.FormComponent, {:saved, post}}, socket) do\n  {:noreply,\n    socket\n    |> subscribe_entity(\"post_with_comments_counts\", :direct, %{\"id\" => post.id})}\nend\n```\n\nThis function gets called when a `posts` has been created.\nAnd inside it, instead of redirecting immediately, we call the `subscribe_entity/4` function to listen for `post_with_comments_counts` with a specific ID. The ID of the newly created post.\n\nThat means that as soon as this `post_with_comments_counts` entity will be created, we will be notified:\n\n```elixir\n@impl true\ndef handle_info({:on_insert, :direct, \"post_with_comments_counts\", post}, socket) do\n  {:noreply,\n    socket\n    |> unsubscribe_entity(\"post_with_comments_counts\", :direct)\n    |> push_navigate(to: \"/posts/#{post.new_record.id}\")}\nend\n```\n\nWhen the entity is created, we unsubscribe from the channel (this is optional is it will automatically unsubscribe as soon as the live view terminates), and we redirect to the detail page.\n\nAnd we know, at this moment, the `post_with_comments_counts` entity will exist.","ref":"readme.html#wait-for-the-entity-to-be-created-before-redirecting"},{"type":"extras","title":"Handle entity not found - EventStreamex","doc":"Another solution is to redirect immediately without waiting for the `post_with_comments_counts` to be created, and by handling the case where the entity is not found yet.\n\n```elixir\ndefmodule MyApp.PostLive.Show do\n  use MyApp, :live_view\n\n  use EventStreamex.EventListener,\n    schema: \"post_with_comments_counts\",\n    subscriptions: [:direct]\n\n  alias MyApp.Blog\n\n  @impl true\n  def mount(params, session, socket) do\n    super(params, session, socket)\n  end\n\n  @impl true\n  def handle_params(%{\"id\" => id} = params, url, socket) do\n    {_res, socket} = super(params, url, socket)\n\n    post = Blog.get_post(id)\n\n    {:noreply,\n     socket\n     |> assign(:id, id)\n     |> assign(:post, post)\n  end\n\n  @impl true\n  def handle_info({:on_insert, :direct, \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> assign(:post, post.new_record)}\n  end\n\n  @impl true\n  def handle_info({:on_update, :direct, \"post_with_comments_counts\", post}, socket) do\n    {:noreply,\n     socket\n     |> assign(:post, post.new_record)}\n  end\n\n  @impl true\n  def handle_info({:on_delete, :direct, \"post_with_comments_counts\", _post}, socket) do\n    {:noreply, socket |> put_flash(:warning, \"Ce post vient d'être supprimé\")}\n  end\nend\n```\n\nHere we fecth the entity (but without triggering an exception if it does not exist), and we replace it if we receive the `:on_insert` event for that entity.\n\nYou will just have to ensure your view shows a loader or something until the entity is created.","ref":"readme.html#handle-entity-not-found"},{"type":"extras","title":"Handling async with APIs - EventStreamex","doc":"We saw 2 solutions we could use with a live view.\nBut what if you develop an mobile app and use an API?\n\nIf you have an API endpoint to create a new `posts` (`POST /posts`), and as soon as your appliation calls it, it redirects to another view requesting the detail of the created `post_with_comments_counts` (`GET /post_with_comments_counts`), then, you may receive a 404 error because of the async creation of the `post_with_comments_counts` entity.\n\nIn order to avoid that, you could implement a mecanism that would return, along with the `posts` newly created, the post creation timestamp.\n\nThen, in the `GET /post_with_comments_counts` endpoint, you would send along the timestamp you got from the `POST /posts` endpoint.\n\nAnd in your controller you would check if the entity has been processed before returning the `post_with_comments_counts` requested.\n\nYou can use `EventStreamex.Operators.ProcessStatus.processed?/2` in order to do so.\n\nFor instance, when you create a `posts`, you receive this header from the API:\n\n```\nX-Event-Streamex-Entity-Timestamp: \"posts/1422057007123000\"\n```\n\n```elixir\ndef create(conn, %{\"post\" => post_params}) do\n  with {:ok, %Post{} = post} <- Blog.create_post(post_params) do\n    conn\n    |> put_status(:created)\n    |> put_resp_header(\"location\", ~p\"/api/posts/#{post}\")\n    |> put_resp_header(\n      \"x-event-streamex-entity-timestamp\",\n      \"posts/#{EventStreamex.Operators.EntityProcessed.to_unix_timestamp!(post.updated_at)}\"\n    )\n    |> render(:show, post: post)\n  end\nend\n```\n\n_(You can use the `:process_status_entity_field` config parameter to set the field to use in your entities to know the last updated time. You should return the field's value in your API, and use this value in the `processed?/2` function)_\n\nYou will send this header when you call the `GET /post_with_comments_counts` endpoint.\n\nAnd the controller will immediately check if the entity has been processed:\n\n```elixir\nEventStreamex.Operators.ProcessStatus.processed?(\"posts\", 1422057007123000)\n```\n\nIf the entity has not been processed yet, then, you know the `post_with_comments_counts` entity has not been created yet (it works the same for updated entities).\nSo you can immediately return a 425 response for instance, informing the client to retry later.\n\nIf the entity has been processed, then you fetch it and return it.\nIf you get a 404 error, this time, that means the entity did not exist.\n\nYou can do that in a plug in your api scope:\n\n```elixir\ndefmodule MyApp.Plugs.EventStreamexEntityTimestamp do\n  import Plug.Conn\n\n  @header_name \"x-event-streamex-entity-timestamp\"\n\n  def init(default), do: default\n\n  def call(%Plug.Conn{} = conn, _default) do\n    case get_req_header(conn, @header_name) do\n      [] ->\n        conn\n\n      [entity_timestamp] ->\n        case decode_entity_timestamp(entity_timestamp) do\n          {:ok, entity, timestamp} ->\n            unless EventStreamex.Operators.ProcessStatus.processed?(entity, timestamp) do\n              conn\n              |> put_status(425)\n              |> resp(425, \"Entity not processed yet\")\n              |> halt()\n            else\n              conn\n            end\n\n          _ ->\n            conn\n        end\n    end\n  end\n\n  def decode_entity_timestamp(entity_timestamp) do\n    case entity_timestamp |> String.split(\"/\") do\n      [entity | [timestamp_str | []]] ->\n        case Integer.parse(timestamp_str, 10) do\n          {timestamp, _} ->\n            {:ok, entity, timestamp}\n\n          :error ->\n            {:error, :not_integer}\n        end\n\n      _ ->\n        {:error, :bad_format}\n    end\n  end\nend\n```\n\nOf course you can implement it the way you want, but you've got the idea.","ref":"readme.html#handling-async-with-apis"},{"type":"extras","title":"Tests - EventStreamex","doc":"To test the package you can run:\n\n```\nmix test\n```\n\nSome tests will need access to a PostgreSQL database.\n\nHere are the credentials used by the tests:\n\n- Database: `postgres`\n- Hostname: `locahost`\n- Password: `\"\"`\n- Port: `5432`\n- Username: `postgres`","ref":"readme.html#tests"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,49]}}